# flake8: noqa:

import bmy
import unittest

import os
from myrrh.utils.myrrh_test_init import *

setUp(f"bmy.{os.path.basename(__file__)}")

with bmy.select():
    from mlib.py import glob
    from mlib.py import os
    from mlib.py import shutil
    from mlib.py import sys

with bmy.select():
    from mlib.py.test.support.os_helper import (
        TESTFN,
        skip_unless_symlink,
        can_symlink,
        create_empty_file,
        change_cwd,
    )


class GlobTests(unittest.TestCase):
    dir_fd = None

    def norm(self, *parts):
        return os.path.normpath(os.path.join(self.tempdir, *parts))

    def joins(self, *tuples):
        return [os.path.join(self.tempdir, *parts) for parts in tuples]

    def mktemp(self, *parts):
        filename = self.norm(*parts)
        base, file = os.path.split(filename)
        if not os.path.exists(base):
            os.makedirs(base)
        create_empty_file(filename)

    def setUp(self):
        self.tempdir = TESTFN + "_dir"
        self.mktemp("a", "D")
        self.mktemp("aab", "F")
        self.mktemp(".aa", "G")
        self.mktemp(".bb", "H")
        self.mktemp(".bb", ".J")
        self.mktemp("aaa", "zzzF")
        self.mktemp("ZZZ")
        self.mktemp("EF")
        self.mktemp("a", "bcd", "EF")
        self.mktemp("a", "bcd", "efg", "ha")
        if can_symlink():
            os.symlink(self.norm("broken"), self.norm("sym1"))
            os.symlink("broken", self.norm("sym2"))
            os.symlink(os.path.join("a", "bcd"), self.norm("sym3"))
        if {os.open, os.stat} <= os.supports_dir_fd and os.scandir in os.supports_fd:
            self.dir_fd = os.open(self.tempdir, os.O_RDONLY | os.O_DIRECTORY)
        else:
            self.dir_fd = None

    def tearDown(self):
        if self.dir_fd is not None:
            os.close(self.dir_fd)
        shutil.rmtree(self.tempdir)

    def glob(self, *parts, **kwargs):
        if len(parts) == 1:
            pattern = parts[0]
        else:
            pattern = os.path.join(*parts)
        p = os.path.join(self.tempdir, pattern)
        res = glob.glob(p, **kwargs)
        res2 = glob.iglob(p, **kwargs)
        self.assertCountEqual(glob.iglob(p, **kwargs), res)

        bres = [os.fsencode(x) for x in res]
        self.assertCountEqual(glob.glob(os.fsencode(p), **kwargs), bres)
        self.assertCountEqual(glob.iglob(os.fsencode(p), **kwargs), bres)

        with change_cwd(self.tempdir):
            res2 = glob.glob(pattern, **kwargs)
            for x in res2:
                self.assertFalse(os.path.isabs(x), x)
            if pattern == "**" or pattern == "**" + os.sep:
                expected = res[1:]
            else:
                expected = res
            self.assertCountEqual([os.path.join(self.tempdir, x) for x in res2], expected)
            self.assertCountEqual(glob.iglob(pattern, **kwargs), res2)
            bpattern = os.fsencode(pattern)
            bres2 = [os.fsencode(x) for x in res2]
            self.assertCountEqual(glob.glob(bpattern, **kwargs), bres2)
            self.assertCountEqual(glob.iglob(bpattern, **kwargs), bres2)

        self.assertCountEqual(glob.glob(pattern, root_dir=self.tempdir, **kwargs), res2)
        self.assertCountEqual(glob.iglob(pattern, root_dir=self.tempdir, **kwargs), res2)
        btempdir = os.fsencode(self.tempdir)
        self.assertCountEqual(glob.glob(bpattern, root_dir=btempdir, **kwargs), bres2)
        self.assertCountEqual(glob.iglob(bpattern, root_dir=btempdir, **kwargs), bres2)

        if self.dir_fd is not None:
            self.assertCountEqual(glob.glob(pattern, dir_fd=self.dir_fd, **kwargs), res2)
            self.assertCountEqual(glob.iglob(pattern, dir_fd=self.dir_fd, **kwargs), res2)
            self.assertCountEqual(glob.glob(bpattern, dir_fd=self.dir_fd, **kwargs), bres2)
            self.assertCountEqual(glob.iglob(bpattern, dir_fd=self.dir_fd, **kwargs), bres2)

        return res

    def assertSequencesEqual_noorder(self, l1, l2):
        l1 = list(l1)
        l2 = list(l2)
        self.assertEqual(set(l1), set(l2))
        self.assertEqual(sorted(l1), sorted(l2))

    def test_glob_literal(self):
        eq = self.assertSequencesEqual_noorder
        eq(self.glob("a"), [self.norm("a")])
        eq(self.glob("a", "D"), [self.norm("a", "D")])
        eq(self.glob("aab"), [self.norm("aab")])
        eq(self.glob("zymurgy"), [])

        res = glob.glob("*")
        self.assertEqual({type(r) for r in res}, {str})
        res = glob.glob(os.path.join(os.curdir, "*"))
        self.assertEqual({type(r) for r in res}, {str})

        res = glob.glob(b"*")
        self.assertEqual({type(r) for r in res}, {bytes})
        res = glob.glob(os.path.join(os.fsencode(os.curdir), b"*"))
        self.assertEqual({type(r) for r in res}, {bytes})

    def test_glob_empty_pattern(self):
        self.assertEqual(glob.glob(""), [])
        self.assertEqual(glob.glob(b""), [])
        self.assertEqual(glob.glob("", root_dir=self.tempdir), [])
        self.assertEqual(glob.glob(b"", root_dir=os.fsencode(self.tempdir)), [])
        self.assertEqual(glob.glob("", dir_fd=self.dir_fd), [])
        self.assertEqual(glob.glob(b"", dir_fd=self.dir_fd), [])

    def test_glob_one_directory(self):
        eq = self.assertSequencesEqual_noorder
        eq(self.glob("a*"), map(self.norm, ["a", "aab", "aaa"]))
        eq(self.glob("*a"), map(self.norm, ["a", "aaa"]))
        eq(self.glob(".*"), map(self.norm, [".aa", ".bb"]))
        eq(self.glob("?aa"), map(self.norm, ["aaa"]))
        eq(self.glob("aa?"), map(self.norm, ["aaa", "aab"]))
        eq(self.glob("aa[ab]"), map(self.norm, ["aaa", "aab"]))
        eq(self.glob("*q"), [])

    def test_glob_nested_directory(self):
        eq = self.assertSequencesEqual_noorder
        if os.path.normcase("abCD") == "abCD":
            # case-sensitive filesystem
            eq(self.glob("a", "bcd", "E*"), [self.norm("a", "bcd", "EF")])
        else:
            # case insensitive filesystem
            eq(
                self.glob("a", "bcd", "E*"),
                [self.norm("a", "bcd", "EF"), self.norm("a", "bcd", "efg")],
            )
        eq(self.glob("a", "bcd", "*g"), [self.norm("a", "bcd", "efg")])

    def test_glob_directory_names(self):
        eq = self.assertSequencesEqual_noorder
        eq(self.glob("*", "D"), [self.norm("a", "D")])
        eq(self.glob("*", "*a"), [])
        eq(self.glob("a", "*", "*", "*a"), [self.norm("a", "bcd", "efg", "ha")])
        eq(self.glob("?a?", "*F"), [self.norm("aaa", "zzzF"), self.norm("aab", "F")])

    def test_glob_directory_with_trailing_slash(self):
        # Patterns ending with a slash shouldn't match non-dirs
        res = glob.glob(self.norm("Z*Z") + os.sep)
        self.assertEqual(res, [])
        res = glob.glob(self.norm("ZZZ") + os.sep)
        self.assertEqual(res, [])
        # When there is a wildcard pattern which ends with os.sep, glob()
        # doesn't blow up.
        res = glob.glob(self.norm("aa*") + os.sep)
        self.assertEqual(len(res), 2)
        # either of these results is reasonable
        self.assertIn(
            set(res),
            [
                {self.norm("aaa"), self.norm("aab")},
                {self.norm("aaa") + os.sep, self.norm("aab") + os.sep},
            ],
        )

    def test_glob_bytes_directory_with_trailing_slash(self):
        # Same as test_glob_directory_with_trailing_slash, but with a
        # bytes argument.
        res = glob.glob(os.fsencode(self.norm("Z*Z") + os.sep))
        self.assertEqual(res, [])
        res = glob.glob(os.fsencode(self.norm("ZZZ") + os.sep))
        self.assertEqual(res, [])
        res = glob.glob(os.fsencode(self.norm("aa*") + os.sep))
        self.assertEqual(len(res), 2)
        # either of these results is reasonable
        self.assertIn(
            set(res),
            [
                {os.fsencode(self.norm("aaa")), os.fsencode(self.norm("aab"))},
                {
                    os.fsencode(self.norm("aaa") + os.sep),
                    os.fsencode(self.norm("aab") + os.sep),
                },
            ],
        )

    @skip_unless_symlink
    def test_glob_symlinks(self):
        eq = self.assertSequencesEqual_noorder
        eq(self.glob("sym3"), [self.norm("sym3")])
        eq(self.glob("sym3", "*"), [self.norm("sym3", "EF"), self.norm("sym3", "efg")])
        self.assertIn(
            self.glob("sym3" + os.sep),
            [[self.norm("sym3")], [self.norm("sym3") + os.sep]],
        )
        eq(
            self.glob("*", "*F"),
            [self.norm("aaa", "zzzF"), self.norm("aab", "F"), self.norm("sym3", "EF")],
        )

    @skip_unless_symlink
    def test_glob_broken_symlinks(self):
        eq = self.assertSequencesEqual_noorder
        eq(self.glob("sym*"), [self.norm("sym1"), self.norm("sym2"), self.norm("sym3")])
        eq(self.glob("sym1"), [self.norm("sym1")])
        eq(self.glob("sym2"), [self.norm("sym2")])

    @unittest.skipUnless(sys.platform == "win32", "Win32 specific test")
    def test_glob_magic_in_drive(self):
        eq = self.assertSequencesEqual_noorder
        eq(glob.glob("*:"), [])
        eq(glob.glob(b"*:"), [])
        eq(glob.glob("?:"), [])
        eq(glob.glob(b"?:"), [])
        eq(glob.glob("\\\\?\\c:\\"), ["\\\\?\\c:\\"])
        eq(glob.glob(b"\\\\?\\c:\\"), [b"\\\\?\\c:\\"])
        eq(glob.glob("\\\\*\\*\\"), [])
        eq(glob.glob(b"\\\\*\\*\\"), [])

    def check_escape(self, arg, expected):
        self.assertEqual(glob.escape(arg), expected)
        self.assertEqual(glob.escape(os.fsencode(arg)), os.fsencode(expected))

    def test_escape(self):
        check = self.check_escape
        check("abc", "abc")
        check("[", "[[]")
        check("?", "[?]")
        check("*", "[*]")
        check("[[_/*?*/_]]", "[[][[]_/[*][?][*]/_]]")
        check("/[[_/*?*/_]]/", "/[[][[]_/[*][?][*]/_]]/")

    @unittest.skipUnless(sys.platform == "win32", "Win32 specific test")
    def test_escape_windows(self):
        check = self.check_escape
        check("?:?", "?:[?]")
        check("*:*", "*:[*]")
        check(r"\\?\c:\?", r"\\?\c:\[?]")
        check(r"\\*\*\*", r"\\*\*\[*]")
        check("//?/c:/?", "//?/c:/[?]")
        check("//*/*/*", "//*/*/[*]")

    def rglob(self, *parts, **kwargs):
        return self.glob(*parts, recursive=True, **kwargs)

    def hglob(self, *parts, **kwargs):
        return self.glob(*parts, include_hidden=True, **kwargs)

    def test_hidden_glob(self):
        eq = self.assertSequencesEqual_noorder
        l = [("aaa",), (".aa",)]
        eq(self.hglob("?aa"), self.joins(*l))
        eq(self.hglob("*aa"), self.joins(*l))
        l2 = [
            (
                ".aa",
                "G",
            )
        ]
        eq(self.hglob("**", "G"), self.joins(*l2))

    def test_recursive_glob(self):
        eq = self.assertSequencesEqual_noorder
        full = [
            ("EF",),
            ("ZZZ",),
            ("a",),
            ("a", "D"),
            ("a", "bcd"),
            ("a", "bcd", "EF"),
            ("a", "bcd", "efg"),
            ("a", "bcd", "efg", "ha"),
            ("aaa",),
            ("aaa", "zzzF"),
            ("aab",),
            ("aab", "F"),
        ]
        if can_symlink():
            full += [
                ("sym1",),
                ("sym2",),
                ("sym3",),
                ("sym3", "EF"),
                ("sym3", "efg"),
                ("sym3", "efg", "ha"),
            ]
        eq(self.rglob("**"), self.joins(("",), *full))
        eq(
            self.rglob(os.curdir, "**"),
            self.joins((os.curdir, ""), *((os.curdir,) + i for i in full)),
        )
        dirs = [
            ("a", ""),
            ("a", "bcd", ""),
            ("a", "bcd", "efg", ""),
            ("aaa", ""),
            ("aab", ""),
        ]
        if can_symlink():
            dirs += [("sym3", ""), ("sym3", "efg", "")]
        eq(self.rglob("**", ""), self.joins(("",), *dirs))

        eq(
            self.rglob("a", "**"),
            self.joins(
                ("a", ""),
                ("a", "D"),
                ("a", "bcd"),
                ("a", "bcd", "EF"),
                ("a", "bcd", "efg"),
                ("a", "bcd", "efg", "ha"),
            ),
        )
        eq(self.rglob("a**"), self.joins(("a",), ("aaa",), ("aab",)))
        expect = [("a", "bcd", "EF"), ("EF",)]
        if can_symlink():
            expect += [("sym3", "EF")]
        eq(self.rglob("**", "EF"), self.joins(*expect))
        expect = [("a", "bcd", "EF"), ("aaa", "zzzF"), ("aab", "F"), ("EF",)]
        if can_symlink():
            expect += [("sym3", "EF")]
        eq(self.rglob("**", "*F"), self.joins(*expect))
        eq(self.rglob("**", "*F", ""), [])
        eq(
            self.rglob("**", "bcd", "*"),
            self.joins(("a", "bcd", "EF"), ("a", "bcd", "efg")),
        )
        eq(self.rglob("a", "**", "bcd"), self.joins(("a", "bcd")))

        with change_cwd(self.tempdir):
            join = os.path.join
            eq(glob.glob("**", recursive=True), [join(*i) for i in full])
            eq(glob.glob(join("**", ""), recursive=True), [join(*i) for i in dirs])
            eq(glob.glob(join("**", "*"), recursive=True), [join(*i) for i in full])
            eq(
                glob.glob(join(os.curdir, "**"), recursive=True),
                [join(os.curdir, "")] + [join(os.curdir, *i) for i in full],
            )
            eq(
                glob.glob(join(os.curdir, "**", ""), recursive=True),
                [join(os.curdir, "")] + [join(os.curdir, *i) for i in dirs],
            )
            eq(
                glob.glob(join(os.curdir, "**", "*"), recursive=True),
                [join(os.curdir, *i) for i in full],
            )
            eq(glob.glob(join("**", "zz*F"), recursive=True), [join("aaa", "zzzF")])
            eq(glob.glob("**zz*F", recursive=True), [])
            expect = [join("a", "bcd", "EF"), "EF"]
            if can_symlink():
                expect += [join("sym3", "EF")]
            eq(glob.glob(join("**", "EF"), recursive=True), expect)

            rec = [(".bb", "H"), (".bb", ".J"), (".aa", "G"), (".aa",), (".bb",)]
            eq(
                glob.glob("**", recursive=True, include_hidden=True),
                [join(*i) for i in full + rec],
            )

    def test_glob_many_open_files(self):
        depth = 30
        base = os.path.join(self.tempdir, "deep")
        p = os.path.join(base, *(["d"] * depth))
        os.makedirs(p)
        pattern = os.path.join(base, *(["*"] * depth))
        iters = [glob.iglob(pattern, recursive=True) for j in range(100)]
        for it in iters:
            self.assertEqual(next(it), p)
        pattern = os.path.join(base, "**", "d")
        iters = [glob.iglob(pattern, recursive=True) for j in range(100)]
        p = base
        for i in range(depth):
            p = os.path.join(p, "d")
            for it in iters:
                self.assertEqual(next(it), p)


@skip_unless_symlink
class SymlinkLoopGlobTests(unittest.TestCase):
    def test_selflink(self):
        tempdir = TESTFN + "_dir"
        os.makedirs(tempdir)
        self.addCleanup(shutil.rmtree, tempdir)
        with change_cwd(tempdir):
            os.makedirs("dir")
            create_empty_file(os.path.join("dir", "file"))
            os.symlink(os.curdir, os.path.join("dir", "link"))

            results = glob.glob("**", recursive=True)
            self.assertEqual(len(results), len(set(results)))
            results = set(results)
            depth = 0
            while results:
                path = os.path.join(*(["dir"] + ["link"] * depth))
                self.assertIn(path, results)
                results.remove(path)
                if not results:
                    break
                path = os.path.join(path, "file")
                self.assertIn(path, results)
                results.remove(path)
                depth += 1

            results = glob.glob(os.path.join("**", "file"), recursive=True)
            self.assertEqual(len(results), len(set(results)))
            results = set(results)
            depth = 0
            while results:
                path = os.path.join(*(["dir"] + ["link"] * depth + ["file"]))
                self.assertIn(path, results)
                results.remove(path)
                depth += 1

            results = glob.glob(os.path.join("**", ""), recursive=True)
            self.assertEqual(len(results), len(set(results)))
            results = set(results)
            depth = 0
            while results:
                path = os.path.join(*(["dir"] + ["link"] * depth + [""]))
                self.assertIn(path, results)
                results.remove(path)
                depth += 1


if __name__ == "__main__":
    unittest.main()



from myrrh.framework.mpython._mosprocess import AbcOsProcess

__mlib__ = "OsProcess"


class OsProcess(AbcOsProcess):
    SIGKILL = 9
    WNOHANG = 1

    def WTERMSIG(self, status: int) -> int:
        return status & 0x7F

    def WIFSTOPPED(self, status: int) -> bool:
        return (status & 0xFF) == 0x7F

    def WSTOPSIG(self, status: int) -> int:
        return (status & 0xFF00) >> 8



import os
import errno
import genericpath
import stat
import typing

import warnings
import functools

from os import PathLike
from collections import namedtuple


from myrrh.core.interfaces import abstractmethod
from myrrh.core.services.system import (
    MOsError,
    MIOException,
    AbcRuntime,
    ImplPropertyClass,
)

from . import mimportlib

__mlib__ = "AbcOsFs"

_stat_named_tuple_field_names = (
    "st_mode",
    "st_ino",
    "st_dev",
    "st_nlink",
    "st_uid",
    "st_gid",
    "st_size",
    "st_atime",
    "st_mtime",
    "st_ctime",
)
_stat_named_tuple_kfield_names = ("st_atime_ns", "st_ctime_ns", "st_mtime_ns")

StatNamedTuple = typing.NamedTuple(  # type: ignore[name-match]
    "stat_result",
    (
        ("st_mode", int),
        ("st_ino", int),
        ("st_dev", int),
        ("st_nlink", int),
        ("st_uid", int),
        ("st_gid", int),
        ("st_size", int),
        ("st_atime", int),
        ("st_mtime", int),
        ("st_ctime", int),
    ),
)


class stat_result(StatNamedTuple):
    def __new__(cls, *fields, **kfields):
        fields_extra = {}

        if len(fields) == 2:
            kfields = fields[1]
            fields = fields[0]

        for k, v in kfields.items():
            if k not in StatNamedTuple._fields:
                fields_extra[k] = v

        if fields:
            stat = super().__new__(cls, *fields[:10])

        elif kfields:
            stat = super().__new__(
                cls,
                **dict(
                    filter(
                        lambda item: item[0] in _stat_named_tuple_field_names,
                        kfields.items(),
                    )
                ),
            )

        for k, v in fields_extra.items():
            super().__setattr__(stat, k, v)

        return stat

    def __setattr__(self, k, v):
        raise AttributeError


statvfs_result = namedtuple(
    "statvfs_result",
    "f_bsize f_frsize f_blocks f_bfree f_bavail f_files f_ffree f_favail f_flag f_namemax",
)


class AbcOsFs(AbcRuntime):
    __frameworkpath__ = "mpython._mosfs"

    __all__ = [
        "mkdir",
        "rmdir",
        "makedirs",
        "unlink",
        "removedirs",
        "exits",
        "lexists",
        "isfile",
        "isdir",
        "islink",
        "getsize",
        "getmtime",
        "getatime",
        "getctime",
        "abspath",
        "expanduser",
        "relpath",
        "splitext",
        "walk",
        "access",
        "chdir",
        "getcwd",
        "listdir",
        "rename",
        "stat",
        "lstat",
        "chown",
        "chmod",
        "normcase",
        "isabs",
        "join",
        "splitdrice",
        "split",
        "basename",
        "dirname",
        "ismount",
        "normpath",
        "realpath",
        "utime",
        "curdir",
        "pardir",
        "extsep",
        "sep",
        "pathsep",
        "altsep",
        "defpath",
        "devnull",
        "scandir",
    ]

    default_mode = -1

    statvfs_result = statvfs_result

    from os import R_OK, X_OK, W_OK, F_OK

    from genericpath import commonprefix  # type: ignore[misc]

    commonprefix = staticmethod(commonprefix)

    @property
    def altsep(self):
        return self.myrrh_os.fsdecode(self.myrrh_os.altsepb)

    @property
    def curdir(self):
        return self.myrrh_os.fsdecode(self.myrrh_os.curdirb)

    @property
    def pardir(self):
        return self.myrrh_os.fsdecode(self.myrrh_os.pardirb)

    @property
    def sep(self):
        return self.myrrh_os.fsdecode(self.myrrh_os.sepb)

    @property
    def extsep(self):
        return self.myrrh_os.fsdecode(self.myrrh_os.extsepb)

    @property
    def pathsep(self):
        return self.myrrh_os.fsdecode(self.myrrh_os.pathsepb)

    @property
    def defpath(self):
        return self.myrrh_os.fsdecode(self.myrrh_os.defpathb)

    @defpath.setter
    def defpath(self, value):
        self.myrrh_os.defpathb = self.myrrh_os.fsencode(value)

    @property
    def devnull(self):
        return self.myrrh_os.fsdecode(self.myrrh_os.devnullb)

    def fspath(self, *a, **kwa):
        return self.myrrh_os._fspath_(*a, **kwa)

    class DirEntry(PathLike):
        __path__: str | None = None

        def __fspath__(self):
            return self.__path__

        @functools.cached_property
        def __lstat__(self):
            return self.lstat(self.__path__)

        def __stat__(self, follow_symlinks=True):
            return self.__fstat__ if follow_symlinks else self.__lstat__

        def __init__(self, name, fspath, stat, lstatmethod):
            self.__name__ = name
            self.__path__ = fspath
            self.__fstat__ = stat
            self.lstat = lstatmethod

        def __repr__(self):
            if isinstance(self.__name__, bytes):
                return b"<DirEntry '%s'>" % self.__name__
            return "<DirEntry '%s'>" % self.__name__

        @property
        def name(self):
            return self.__name__

        @property
        def path(self):
            return self.__path__

        def inode(self):
            return self.__fstat__.st_ino

        def is_dir(self, *, follow_symlinks=True):
            return stat.S_ISDIR(self.__stat__(follow_symlinks).st_mode)

        def is_file(self, *, follow_symlinks=True):
            return stat.S_ISREG(self.__stat__(follow_symlinks).st_mode)

        def is_symlink(self):
            try:
                return stat.S_ISLNK(self.__lstat__.st_mode)
            except FileNotFoundError:
                pass

            return False

        def stat(self, *, follow_symlinks=True):
            return self.__stat__(follow_symlinks)

    _os = mimportlib.module_property("_osenv")

    def mkdir(self, path, mode=0o777):
        path = self.myrrh_os.p(path)

        if path.endswith((self.myrrh_os.sepb, b"/")):
            path = path[:-1]

        if self.myrrh_os.fs.exist(path):
            MOsError(self, errno=errno.EEXIST, args=(path,)).raised()

        dirname = self.dirname(path)

        if dirname != b"" and not self.isdir(dirname):
            MOsError(self, errno=errno.ENOENT, args=(dirname,)).raised()

        try:
            self.myrrh_os.fs.mkdir(path)
        except Exception as e:
            MIOException(self, strerror="operation failed", exc=e, filename=path).raised()

        self.chmod(path, mode, _trusted=True)

    def rmdir(self, path):
        _path = self.myrrh_os.p(path)
        try:
            self.myrrh_os.fs.rmdir(_path)
        except UnicodeDecodeError:
            raise
        except Exception as e:
            if not self.isdir(path):
                MOsError(self, errno.ENOTDIR, "Not a directory", args=(path,)).raised()

            MIOException(self, strerror="operation failed", exc=e, filename=path).raised()

    def makedirs(self, path, mode=0o777, exist_ok=False):
        path = self.myrrh_os.p(path)

        if not exist_ok and self.myrrh_os.fs.exist(path):
            MOsError(self, errno=errno.EEXIST).raised()
        try:
            self.myrrh_os.fs.mkdir(path)
        except Exception as e:
            MIOException(self, strerror="operation failed", exc=e, filename=path).raised()

        try:
            self.chmod(path, mode, _trusted=True)
        except PermissionError:
            pass

    def unlink(self, path, *, dir_fd=None):
        return self.remove(path, dir_fd=dir_fd)

    def remove(self, path, *, dir_fd=None):
        _path = self.myrrh_os.p(path, dir_fd=dir_fd)
        try:
            self.myrrh_os.fs.rm(_path)
        except UnicodeDecodeError:
            raise
        except Exception as e:
            MIOException(self, strerror="operation failed", exc=e, filename=path).raised()

    def removedirs(self, name):
        try:
            rmdir = self.myrrh_os.fs.rmdir
            path = self
            rmdir(self.myrrh_os.p(name))
            head, tail = path.split(name)
            if not tail:
                head, tail = path.split(head)
            while head and tail:
                try:
                    rmdir(self.myrrh_os.p(head))
                except Exception:
                    break
                head, tail = path.split(head)
        except Exception as e:
            MIOException(self, strerror="operation failed", exc=e, filename=path).raised()

    def exists(self, path):
        path = self.myrrh_os.p(path)
        return len(path) != 0 and self.myrrh_os.fs.exist(path)

    def lexists(self, path):
        try:
            self.lstat(path)
        except (MOsError, FileNotFoundError):
            return False

        return True

    def isfile(self, path):
        path = self.myrrh_os.p(path)

        if len(path) == 0 or not self.myrrh_os.fs.exist(path):
            return False

        try:
            st = self.stat(path)
        except (MOsError, FileNotFoundError):
            return False

        return stat.S_ISREG(st.st_mode)

    def isdir(self, path):
        path = self.myrrh_os.p(path)

        if len(path) == 0 or not self.myrrh_os.fs.exist(path):
            return False

        isdir = self.myrrh_os.fs.is_container(path)

        if not isdir:
            try:
                st = self.stat(path)
            except (MOsError, FileNotFoundError):
                return False

            isdir = stat.S_ISDIR(st.st_mode)

        return isdir

    def islink(self, path):
        try:
            st = self.lstat(path)
        except (MOsError, FileNotFoundError):
            return False
        return stat.S_ISLNK(st.st_mode)

    def getsize(self, filename):
        filename = self.myrrh_os.p(filename)
        return self.stat(filename).st_size

    def getmtime(self, filename):
        filename = self.myrrh_os.p(filename)
        return self.stat(filename).st_mtime

    def getatime(self, filename):
        filename = self.myrrh_os.p(filename)
        return self.stat(filename).st_atime

    def getctime(self, filename):
        filename = self.myrrh_os.p(filename)
        return self.stat(filename).st_ctime

    def abspath(self, path):
        _cast_ = self.myrrh_os.fdcast(path)
        return _cast_(self.myrrh_os.getpathb(path))

    def expanduser(self, path):
        _cast_ = self.myrrh_os.fdcast(path)
        _path = self.myrrh_os.p(path)
        if not _path.startswith(b"~"):
            return path
        home = self.myrrh_os.p(self._os.gethome())
        _path = _path[2:] if _path.startswith(b"~/") else _path[1:]
        return _cast_(self.join(home, _path))

    @abstractmethod
    def expandvars(self, path):
        ...

    def relpath(self, path, start=None):
        _cast_ = self.myrrh_os.fdcast(path)
        path = self.myrrh_os.p(path)
        start = self.myrrh_os.p(start)
        sep = self.myrrh_os.sepb
        pardir = self.myrrh_os.pardirb
        curdir = self.myrrh_os.curdirb

        if start is None:
            start = curdir

        if not path:
            raise ValueError("no path specified")

        start_list = [x for x in self.abspath(start).split(sep) if x]
        path_list = [x for x in self.abspath(path).split(sep) if x]

        i = len(self.commonprefix([start_list, path_list]))

        rel_list = [pardir] * (len(start_list) - i) + path_list[i:]
        if not rel_list:
            return _cast_(curdir)

        return _cast_(self.join(*rel_list))

    def splitext(self, path):
        _cast_ = self.myrrh_os.fdcast(path)
        return genericpath._splitext(
            path,
            _cast_(self.myrrh_os.sepb),
            _cast_(self.myrrh_os.altsepb),
            _cast_(self.myrrh_os.extsepb),
        )

    def access(self, path, mode):
        path = self.myrrh_os.p(path)
        mode |= self.F_OK  # unused warning disable

        try:
            st_r = self.stat(path)
        except MOsError:
            return False

        uid = self._os.geteuid()
        gids = self._os.getgroups()

        result = (
            (not ((mode & self.R_OK) == self.R_OK) or ((st_r.st_mode & stat.S_IRUSR) == stat.S_IRUSR and (st_r.st_uid == uid)) or ((st_r.st_mode & stat.S_IRGRP) == stat.S_IRGRP and (st_r.st_gid in gids)) or (st_r.st_mode & stat.S_IROTH) == stat.S_IROTH)
            and (not ((mode & self.W_OK) == self.W_OK) or ((st_r.st_mode & stat.S_IWUSR) == stat.S_IWUSR and (st_r.st_uid == uid)) or ((st_r.st_mode & stat.S_IWGRP) == stat.S_IWGRP and (st_r.st_uid in gids)) or (st_r.st_mode & stat.S_IWOTH) == stat.S_IWOTH)
            and (not ((mode & self.X_OK) == self.X_OK) or ((st_r.st_mode & stat.S_IXUSR) == stat.S_IXUSR and (st_r.st_uid == uid)) or ((st_r.st_mode & stat.S_IXGRP) == stat.S_IXGRP and (st_r.st_uid in gids)) or (st_r.st_mode & stat.S_IXOTH) == stat.S_IXOTH)
        )
        return result

    def chdir(self, path):
        _cast_ = self.myrrh_os.fdcast(path)
        bpath = self.myrrh_os.p(path)

        if self.isdir(bpath):
            self.myrrh_os.cwdb = self.abspath(bpath)
            return

        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), _cast_(path))

    def getcwd(self):
        return self.myrrh_os.fsdecode(self.myrrh_os.getpathb())

    def getcwdb(self):
        return self.myrrh_os.getpathb()

    @abstractmethod
    def link(self, src, dst, *, src_dir_fd=None, dst_dir_fd=None, follow_symlinks=True):
        ...

    def isabs(self, path):
        return self.myrrh_os.isabs(path)

    def join(self, *path):
        return self.myrrh_os.joinpath(*path)

    def fsencode(self, filename):
        filename = self.myrrh_os._fspath_(filename)

        _fsencoding = self.myrrh_os.fsencoding
        _fserrors = self.myrrh_os.fsencodeerrors

        if isinstance(filename, bytes):
            return filename
        elif isinstance(filename, (str,)):
            return filename.encode(_fsencoding, _fserrors)
        else:
            raise TypeError("expect bytes or str, not %s" % type(filename).__name__)

    def fsdecode(self, filename):
        filename = self.myrrh_os._fspath_(filename)

        _fsencoding = self.myrrh_os.fsencoding
        _fserrors = self.myrrh_os.fsencodeerrors

        if isinstance(filename, (str,)):
            return filename
        elif isinstance(filename, bytes):
            return filename.decode(_fsencoding, _fserrors)
        else:
            raise TypeError("expect bytes or str, not %s" % type(filename).__name__)

    def normpath(self, path):
        return self.myrrh_os.normpath(path)

    @abstractmethod
    def listdir(self, path="."):
        ...

    @abstractmethod
    def rename(self, src, dst):
        ...

    @abstractmethod
    def replace(self, src, dst, *, src_dir_fd=None, dst_dir_fd=None):
        ...

    @abstractmethod
    def _scandir_list(self, path="."):
        ...

    class scandir(metaclass=ImplPropertyClass):
        def __init__(self, path="."):
            if path == "":
                raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), path)

            self._entries = self._scandir_list(path)
            self._iter = iter(self._entries)

        def __enter__(self):
            return self

        def __exit__(self, type, value, traceback):
            pass

        def __iter__(self):
            return self

        def __next__(self):
            return next(self._iter)

        def close(self):
            if self._entries:
                del self._entries
            self._entries = None

    def stat_float_times(self, _):
        warnings.warn("stat_float_times deprecated => ignored", DeprecationWarning)

    @abstractmethod
    def _stat(self, path, *, dir_fd=None, follow_symlinks=True):
        pass

    def stat(self, path, *, dir_fd=None, follow_symlinks=True):
        _path = self.myrrh_os.f(path, dir_fd=dir_fd)

        if len(_path) == 0:
            MOsError(self, errno.ENOENT, os.strerror(errno.ENOENT), args=(path,)).raised()

        if not self.myrrh_os.fs.exist(_path):
            MOsError(self, errno.ENOENT, os.strerror(errno.ENOENT), args=(path,)).raised()

        return self._stat(path, follow_symlinks=follow_symlinks)

    def fstatvfs(self, fd):
        return self.statvfs(fd)

    @abstractmethod
    def _statvfs(self, path):
        pass

    def statvfs(self, path):
        _path = self.myrrh_os.f(path)

        if len(_path) == 0:
            MOsError(self, errno.ENOENT, os.strerror(errno.ENOENT), args=(path,)).raised()

        if not self.myrrh_os.fs.exist(_path):
            MOsError(self, errno.ENOENT, os.strerror(errno.ENOENT), args=(path,)).raised()

        return self._statvfs(path)

    @abstractmethod
    def _lstat(self, path, *, dir_fd=None):
        pass

    def lstat(self, path, *, dir_fd=None):
        _path = self.myrrh_os.f(path, dir_fd=dir_fd)

        if len(_path) == 0:
            MOsError(self, errno.ENOENT, os.strerror(errno.ENOENT), args=(path,)).raised()

        if not self.myrrh_os.fs.exist(_path):
            MOsError(self, errno.ENOENT, os.strerror(errno.ENOENT), args=(path,)).raised()

        return self._lstat(path, dir_fd=dir_fd)

    @abstractmethod
    def chown(self, path, uid, gid, *, dir_fd=None, follow_symlinks=True, _trusted=False):
        ...

    @abstractmethod
    def chmod(self, path, mode, *, dir_fd=None, follow_symlinks=True, _trusted=False):
        ...

    def lchmod(self, path, mode):
        return self.chmod(path, mode, follow_symlinks=False)

    @abstractmethod
    def normcase(self, path):
        ...

    @abstractmethod
    def splitdrive(self, path):
        ...

    @abstractmethod
    def split(self, path):
        ...

    def basename(self, path):
        return self.myrrh_os.basename(path)

    def dirname(self, path):
        return self.myrrh_os.dirname(path)

    @abstractmethod
    def ismount(self, path):
        ...

    @abstractmethod
    def realpath(self, path):
        ...

    @abstractmethod
    def _utime(self, path, times=None, *, ns=None, dir_fd=None, follow_symlinks=True):
        pass

    def utime(self, path, times=None, *, ns=None, dir_fd=None, follow_symlinks=True):
        if ns is not None and not (isinstance(ns, tuple) or len(ns) != 2):
            raise TypeError("utime: 'ns' must be a tuple of two ints")

        if times is not None and ns is not None:
            raise ValueError("utime: you may specify either 'times' or 'ns' but not both")

        self._utime(path, times=times, ns=ns, dir_fd=dir_fd, follow_symlinks=follow_symlinks)

    @abstractmethod
    def samefile(self, path1, path2):
        ...

    def sameopenfile(self, fp1, fp2):
        return self.samefile(self.myrrh_os.f(fp1), self.myrrh_os.f(fp2))

    @property
    @abstractmethod
    def _umask(self):
        ...

    @abstractmethod
    def umask(self, mask):
        ...

    @abstractmethod
    def truncate(self, path, length):
        ...

    # this function is distributed under the terms of PSF licence.
    # updated to work with myrrh API.
    # original version : see os module source code
    def walk(self, top, topdown=True, onerror=None, followlinks=False):
        scandir = self.scandir
        path = self
        walk = self.walk

        dirs = []
        nondirs = []
        walk_dirs = []

        try:
            top = self.myrrh_os._fspath_(top)
            scandir_it = scandir(top)
        except OSError as error:
            if onerror is not None:
                onerror(error)
            return

        with scandir_it:
            while True:
                try:
                    try:
                        entry = next(scandir_it)
                    except StopIteration:
                        break
                except OSError as error:
                    if onerror is not None:
                        onerror(error)
                    return

                try:
                    is_dir = entry.is_dir()
                except OSError:
                    # If is_dir() raises an OSError, consider that the entry is not
                    # a directory, same behaviour than os.path.isdir().
                    is_dir = False

                if is_dir:
                    dirs.append(entry.name)
                else:
                    nondirs.append(entry.name)

                if not topdown and is_dir:
                    # Bottom-up: recurse into sub-directory, but exclude symlinks to
                    # directories if followlinks is False
                    if followlinks:
                        walk_into = True
                    else:
                        try:
                            is_symlink = entry.is_symlink()
                        except OSError:
                            # If is_symlink() raises an OSError, consider that the
                            # entry is not a symbolic link, same behaviour than
                            # os.path.islink().
                            is_symlink = False
                        walk_into = not is_symlink

                    if walk_into:
                        walk_dirs.append(entry.path)

        # Yield before recursion if going top down
        if topdown:
            yield top, dirs, nondirs

            # Recurse into sub-directories
            islink, join = path.islink, path.join
            for dirname in dirs:
                new_path = join(top, dirname)
                # Issue #23605: os.path.islink() is used instead of caching
                # entry.is_symlink() result during the loop on os.scandir() because
                # the caller can replace the directory entry during the "yield"
                # above.
                if followlinks or not islink(new_path):
                    yield from walk(new_path, topdown, onerror, followlinks)
        else:
            # Recurse into sub-directories
            for new_path in walk_dirs:
                yield from walk(new_path, topdown, onerror, followlinks)
            # Yield after recursion if going bottom up
            yield top, dirs, nondirs



import collections
import json
import typing

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate, ExecutionFailureCauseRVal
from myrrh.utils import mshlex

__mlib__ = "AbcSys"


class _interface(ABC):
    import sys as local_sys

    @property
    @abstractmethod
    def flags(self) -> typing.Any:
        ...

    @property
    @abstractmethod
    def maxsize(self) -> int:
        ...

    @property
    @abstractmethod
    def hash_info(self) -> typing.Any:
        ...

    @property
    @abstractmethod
    def _git(self) -> tuple:
        ...

    @abstractmethod
    def implementation(self) -> tuple:
        ...

    @property
    @abstractmethod
    def stderr(self) -> object:
        ...

    @property
    @abstractmethod
    def stdout(self) -> object:
        ...

    @property
    @abstractmethod
    def stdin(self) -> object:
        ...

    @property
    @abstractmethod
    def __stderr__(self) -> object:
        ...

    @property
    @abstractmethod
    def __stdout__(self) -> object:
        ...

    @property
    @abstractmethod
    def __stdin__(self) -> object:
        ...

    @abstractmethod
    def getsizeof(self, obj: object, default: int = ...) -> int:
        ...

    @abstractmethod
    def exc_info(self) -> tuple:
        ...

    @abstractmethod
    def audit(self, event, *args) -> None:
        ...


class AbcSys(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.msys"

    __all__ = [
        "local_sys",
        "maxsize",
        "prefix",
        "version",
        "path",
        "version_info",
        "base_prefix",
        "exec_prefix",
        "base_exec_prefix",
        "executable",
        "implementation",
        "platform",
    ]

    __version_info_tuple__ = collections.namedtuple("sys_version_info", "major minor micro releaselevel serial")  # type: ignore[name-match]

    attr_list = "('version', 'version_info', 'prefix', 'base_prefix', 'exec_prefix', 'base_exec_prefix', 'path', 'maxsize', 'abiflags')"

    _version = "0.0.0"
    _version_info = [0, 0, 0, "na", 0]
    _prefix = ""
    _base_prefix = ""
    _exec_prefix = ""
    _base_exec_prefix = ""
    _path: list[str] = []
    _maxsize = 2**31

    __plateform: str | None = None

    __delegated__ = {_interface: _interface.local_sys}
    __delegate_check_type__ = False

    def __init__(self, *a, **kwa):
        self.__delegate__(_interface, self.local_sys)

    @property
    def modules(self):
        return self.myrrh_os.__m_runtime_cache__.modules

    def _get_sys_attr(self, attr):
        instance_attr_name = "_%s" % attr

        try:
            executable = self._executable
        except ExecutionFailureCauseRVal:
            executable = b""

        if getattr(self, instance_attr_name) is getattr(self.__class__, instance_attr_name) and executable != b"":
            try:
                out, err, rval = self.myrrh_os.cmdb(
                    b'%s -c "import sys,json; print(json.dumps({n : getattr(sys, n) for n in %s if hasattr(sys, n) }))"'
                    % (
                        mshlex.dquote(executable),
                        self.myrrh_os.shencode(self.attr_list),
                    )
                )
                ExecutionFailureCauseRVal(self, err, rval, 0).check()
            except (ExecutionFailureCauseRVal, OSError):
                pass
            else:
                out = json.loads(self.myrrh_os.shdecode(out))
                for k, v in out.items():
                    setattr(self, "_%s" % k, v)

        return getattr(self, instance_attr_name)

    @property
    def prefix(self):
        return self._get_sys_attr("prefix")

    @property
    def version(self):
        return self._get_sys_attr("version")

    @version.setter
    def version(self, value):
        self._version = value

    @property
    def path(self):
        return self._get_sys_attr("path")

    @property
    def version_info(self):
        return self.__version_info_tuple__(*self._get_sys_attr("version_info"))

    @property
    def base_prefix(self):
        return self._get_sys_attr("base_prefix")

    @property
    def exec_prefix(self):
        return self._get_sys_attr("base_prefix")

    @property
    def base_exec_prefix(self):
        return self._get_sys_attr("base_exec_prefix")

    @property
    def executable(self):
        try:
            executable = self._executable
            if executable:
                return self.myrrh_os.shdecode(self._executable)
        except ExecutionFailureCauseRVal:
            pass

        raise AttributeError("sys.executable not supported on %s" % self.cfg.id) from None

    @property
    def platform(self):
        if self.__plateform is None:
            self.__plateform = self._platform

        return self.__plateform

    @platform.setter
    def platform(self, value):
        self.__plateform = value

    @property
    @abstractmethod
    def _platform(self):
        """This string contains a platform identifier that can be used to append platform-specific warehouse.
        the values are:
            Linux  => 'linux'
            Windows =>'win32'
        """

    @property
    @abstractmethod
    def _executable(self):
        """
        Python executable path
        """

    def getfilesystemencoding(self):
        return self.myrrh_os.fsencoding


setattr(
    AbcSys,
    "_implementation.__dict__",
    {
        "name": b"",
        "version": AbcSys._version,
        "hexversion": 0,
        "cache_tag": b"",
        "_multiarch": b"",
    },
)



import pprint
import pydantic
import typing
import json

from myrrh.warehouse.item import NoneItem
from myrrh.core.services.entity import Entity, CoreProvider
from myrrh.provider import IProvider

from myrrh.provider.registry import ProviderRegistry
from myrrh.warehouse.registry import ItemRegistry


__all__ = ("Assembly",)


class _FactoryModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid", validate_assignment=True, validate_default=True, frozen=True)

    schema_: str = pydantic.Field("http://./myrrh/1.0/schema.json", alias="$schema")
    supply: ItemRegistry().FactorySupply  # type: ignore[valid-type]
    warehouse: list[ItemRegistry().WarehouseItemT] = pydantic.Field(default_factory=list)  # type: ignore[valid-type]

    @pydantic.field_validator("supply")
    def validate_supply(cls, supply, info):
        return supply


class Assembly:
    def __init__(self, supply={}, warehouse=list(), *, _model=None, _providers=None):
        self._providers = _providers

        if _model:
            self._factory = _model
        elif _providers:
            self._factory = _FactoryModel(supply={"paths": ["*"]}, warehouse=warehouse)
        else:
            self._factory = _FactoryModel(supply=supply, warehouse=warehouse)

    @classmethod
    def Warehouse(cls, *a, **kwa):
        return ItemRegistry().WarehouseItemT(*a, **kwa)

    @classmethod
    def Supply(cls, *a, **kwa):
        return ItemRegistry().FactorySupply(*a, **kwa)

    @classmethod
    def provider_name(cls, path: str) -> str:
        _, _, pname = path.rpartition("/")
        return pname

    @classmethod
    def fromProvider(cls, provider: typing.Type[IProvider], *, warehouse=list(), **kwargs):
        supply = ItemRegistry().FactorySupply(
            paths=[f"**/{provider._name_}"],
            settings=[{"name": provider._name_, **kwargs}],
        )
        return cls(supply=supply, warehouse=warehouse)

    @classmethod
    def fromObj(cls, obj):
        return cls(_model=_FactoryModel.model_validate(obj))

    @classmethod
    def fromRaw(cls, data):
        return cls(_model=_FactoryModel.model_validate_json(data))

    @classmethod
    def fromFile(cls, file):
        with open(file) as model:
            data = model.read()

        return cls(_model=_FactoryModel.model_validate_json(data))

    @classmethod
    def schema(cls, *a, **kwa):
        indent = kwa.pop("indent", None)
        return json.dumps(_FactoryModel.model_json_schema(*a, **kwa), indent=indent)

    @property
    def warehouse(self) -> list[ItemRegistry().WarehouseItemT]:  # type: ignore[valid-type]
        return self._factory.warehouse

    @property
    def supply(self) -> ItemRegistry().FactorySupply:  # type: ignore[valid-type]
        return self._factory.supply

    def _new_providers(self):
        if not self._providers:
            self._providers = list()

            for path in self.supply.paths:
                pname = self.provider_name(path)
                setting = next(filter(lambda p: p.name == pname, self.supply.settings), None)
                provider = getattr(ProviderRegistry(), pname)(setting)

                self._providers.append(provider)

        return CoreProvider(self._providers, patterns=self.supply.paths)

    def build(self) -> Entity:
        for prestep in self.supply.pre:
            pass

        provider = self._new_providers()

        entity = Entity(provider, self.warehouse)

        for poststep in self.supply.post:
            pass

        return entity

    def fromEntity(self, entity: Entity, only_predefined=True):
        if only_predefined:
            warehouse = entity.cfg.predefined()
        else:
            warehouse = entity.cfg.values()

        return self.__class__(supply=self.supply, warehouse=warehouse)

    def get_item(self, type_: str):
        t = [item for item in filter(lambda item: item.type_ == type_, self.warehouse)]
        if len(t) == 1:
            return t[0]

        return NoneItem

    def json(
        self,
        *,
        by_alias=True,
        exclude_defaults=True,
        exclude=None,
        exclude_unset=True,
        **kwa,
    ):
        exclude = exclude or {}
        exclude.update({"warehouse": {i for i in range(len(self.warehouse)) if self.warehouse[i].type_ == "null"}})
        return self._factory.model_dump_json(
            by_alias=by_alias,
            exclude=exclude,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            **kwa,
        )

    def pformat(self, *a, **kwa):
        return pprint.pformat(self._factory, *a, **kwa)

    def dict(self, *a, **kwa):
        return self._factory.model_dump(*a, **kwa)



import functools

from myrrh.framework.mpython._mosenv import AbcOsEnv
from myrrh.core.services.system import ExecutionFailureCauseRVal

__mlib__ = "OsEnv"


class OsEnv(AbcOsEnv):
    name = "posix"
    linesep = "\n"

    __all__ = AbcOsEnv.__all__
    __all__.extend(["uname"])

    def __init__(self, *a, **kwa):
        self._path = None

    def putenv(self, varname, value):
        _, err, rval = self.myrrh_os.cmdb(
            b'%(echo)s "export %(varname)s=%(value)s" >> $HOME/.profile',
            varname=self.myrrh_os.sh_escape_bytes(varname),
            value=self.myrrh_os.sh_escape_bytes(value),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        _, err, rval = self.myrrh_os.cmdb(
            b'%(echo)s "export %(varname)s=%(value)s" >> $HOME/.bashrc',
            varname=self.myrrh_os.sh_escape_bytes(varname),
            value=self.myrrh_os.sh_escape_bytes(value),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def unsetenv(self, varname):
        _, err, rval = self.myrrh_os.cmdb(
            rb'%(sed)s -i /"%(varname)s=.*"/d $HOME/.profile',
            varname=self.myrrh_os.sh_escape_bytes(varname),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        _, err, rval = self.myrrh_os.cmdb(
            rb'%(sed)s -i /"%(varname)s=.*"/d $HOME/.bashrc',
            varname=self.myrrh_os.sh_escape_bytes(varname),
        )
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

    def gethome(self):
        out, err, rval = self.myrrh_os.cmd(b"%(echo)s $HOME")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out.strip()

    def gettemp(self):
        out, err, rval = self.myrrh_os.cmd(b"%(dirname)s `mktemp -u`")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out.strip()

    def getlogin(self):
        out, err, rval = self.myrrh_os.cmd(b"%(echo)s $USER")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out.strip()

    def geteuid(self):
        out, err, rval = self.myrrh_os.cmdb(b"%(id)s -u")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return int(out.strip())

    def getegid(self):
        out, err, rval = self.myrrh_os.cmdb(b"%(id)s -g")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return int(out.strip())

    def getuid(self):
        out, err, rval = self.myrrh_os.cmdb(b"%(id)s -u -r")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return int(out.strip())

    def getgid(self):
        out, err, rval = self.myrrh_os.cmdb(b"%(id)s -g -r")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return int(out.strip())

    def getgroups(self):
        out, err, rval = self.myrrh_os.cmdb(b"%(id)s -G")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return [int(g) for g in out.strip().split(b" ")]

    def cpu_count(self):
        out, err, rval = self.myrrh_os.cmdb(b"%(cat)s  /proc/cpuinfo")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        s = 0
        c = 0
        while s > -1:
            c += 1
            s = out.find(b"processor", s + len(b"processor"))
        return None if c == 0 else c

    @functools.cached_property
    def _uname(self):
        out, err, rval = self.myrrh_os.cmd(b"for p in -s -n -r -v -m; do %(uname)s $p; done")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

        infos = out.splitlines()
        if len(infos) != self.uname_result.n_fields:
            return self.uname_result([""] * self.uname_result.n_fields)

        return self.uname_result(infos)

    def uname(self):
        return self._uname



from myrrh.framework.mpython.msys import _interface as Sys
from myrrh.framework.mpython.mgzip import _interface as Gzip
from myrrh.framework.mpython.mfnmatch import _interface as Fnmatch
from myrrh.framework.mpython.mshutil import _interface as Shutil
from myrrh.framework.mpython.mglob import _interface as Glob
from myrrh.framework.mpython.mio import _interface as Io
from myrrh.framework.mpython.msubprocess import _interface as Subprocess
from myrrh.framework.mpython.mplatform import _interface as Platform
from myrrh.framework.mpython.msysconfig import AbcSysConfig
from myrrh.framework.mpython.mtarfile import _interface as Tarfile
from myrrh.framework.mpython.mtempfile import _interface as Tempfile
from myrrh.framework.mpython.mlocale import _interface as Locale

from myrrh.framework.mpython._mosenv import AbcOsEnv
from myrrh.framework.mpython._mosfs import AbcOsFs
from myrrh.framework.mpython._mosfile import AbcOsFile
from myrrh.framework.mpython._mosprocess import AbcOsProcess
from myrrh.framework.mpython.mos import AbcOs


class Os(AbcOsEnv, AbcOsFs, AbcOsFile, AbcOsProcess, AbcOs):
    ...


os: Os
sys: Sys
gzip: Gzip
fnmatch: Fnmatch
shutil: Shutil
glob: Glob
io: Io
subprocess: Subprocess
platform: Platform
sysconfig: AbcSysConfig
tarfile: Tarfile
tempfile: Tempfile
locale: Locale

try:
    from myrrh.framework.arch.posix.mpython.mpwd import AbcPwd

    pwd: AbcPwd
except ImportError:
    ...

try:
    from myrrh.framework.arch.posix.mpython.mgrp import AbcGrp

    grp: AbcGrp
except ImportError:
    ...



# -*- coding: utf-8 -*-

import sys
import threading
import warnings
import os
import signal

# from myrrh.utils import getch

from myrrh.core.services.system import MOsError, AbcRuntime, ImplPropertyClass

from . import mimportlib

# TODO
# from mlib.sh import ansi

__mlib__ = "AbcOsProcess"


class AbcOsProcess(AbcRuntime):
    __frameworkpath__ = "mpython._mosprocess"

    __all__ = [
        "SIGABRT",
        "SIGFPE",
        "SIGINT",
        "SIGILL",
        "SIGTERM",
        "SIGSEGV",
        "P_WAIT",
        "P_NOWAIT",
        "waitpid",
        "spawnv",
        "spawnve",
        "spawnl",
        "spawnle",
        "kill",
        "system",
        "pipe",
    ]

    SIGABRT = signal.SIGABRT
    SIGFPE = signal.SIGFPE
    SIGINT = signal.SIGINT
    SIGILL = signal.SIGILL
    SIGTERM = signal.SIGTERM
    SIGSEGV = signal.SIGSEGV

    P_WAIT = os.P_WAIT
    P_NOWAIT = os.P_NOWAIT
    P_INTERACT = -1

    WNOHANG = P_NOWAIT

    _POLL_READ_TIMEOUT = 0.01
    _POLL_EXIT_TIMEOUT = 0.1

    osfs = mimportlib.module_property("_mosfs")

    def _try_decode(self, out):
        try:
            if self.myrrh_os.defaultencoding():
                out = out.decode(self.myrrh_os.defaultencoding(), errors="ignore")
        except Exception:
            pass

        return out

    def _proc_stdin_listener(self, to_func, stop, chIt, emergency):
        raise NotImplementedError
        """
        getc = keymap(self, chIt)
        while not stop.is_set():
            try:
                c = next(getc)
                if c:
                    to_func(c.encode())
            except KeyboardInterrupt:
                emergency()
            except (StopIteration, BrokenPipeError):
                break
            except Exception as e:
                break
        """

    def _proc_stdout_listener(self, proc):
        out = b""
        err = b""
        cont = True

        while cont:
            try:
                out = proc.output(timeout=self._POLL_READ_TIMEOUT)
                if out is not None:
                    sys.stdout.write(self._try_decode(out))
            except Exception:
                out = None

            try:
                err = proc.error(timeout=self._POLL_READ_TIMEOUT)
                if err is not None:
                    sys.stderr.write(self._try_decode(err))
            except Exception:
                err = None

            cont = out is not None or err is not None or proc.exit_status is None

    def _spawnve(self, mode, path, args, env) -> int:
        raise NotImplementedError

        if mode not in [self.P_NOWAIT, self.P_WAIT, self.P_INTERACT]:
            raise ValueError("invalid mode")
        if not isinstance(args, (list, tuple)):
            raise ValueError("invalid type for arg")
        if args is None or len(args) == 0 or args[0] is None or len(args[0]) == 0:
            raise ValueError("spawn() arg 2 cannot be empty")

        path = self.osfs.myrrh_os.p(path)

        if len(args) > 0 and self.osfs.myrrh_os.p(args[0]) != path:
            warnings.warn("not implemented yet args0 redefinition : arg0 must be equal to path")

        stream = ""

        if mode is self.P_WAIT:
            stream += "r"
        if mode is self.P_INTERACT:
            stream += "w"

        handle = self.myrrh_syscall.openprocess(
            [path] + [self.myrrh_os.fsencode(a) for a in args[1:]],
            stream,
            env=None if env is None else {self.myrrh_os.fsencode(k): self.myrrh_os.fsencode(v) for k, v in env.items()},
        )

        th_out = threading.Thread(
            name="_spawnve_stdout_listener",
            target=self._proc_stdout_listener,
            args=(handle,),
        )
        th_out.daemon = True
        th_out.start()

        if mode is self.P_INTERACT:
            # chIt = getch.getchIt().__enter__()
            stopEvent = threading.Event()
            # th_in = threading.Thread(name="_spawnve_stdin_listener", target=self._proc_stdin_listener, args=(proc.input, stopEvent, chIt, handle.terminate))
            # th_in.daemon = True
            # th_in.start()

        if mode is self.P_WAIT or mode is self.P_INTERACT:
            try:
                _, exit_code = self.waitpid(int(handle))
                th_out.join()
                return exit_code >> 8
            except Exception:
                # try to clean
                try:
                    handle.signal(self.SIGTERM)
                except Exception:
                    pass

            finally:
                if mode is self.P_INTERACT:
                    stopEvent.set()
                    # chIt = getch.getchIt().__exit__()
                    # th_in.join()

        return handle.detach()

    def pipe(self):
        return self.myrrh_syscall.open_pipe()

    def waitpid(self, pid, options=None):
        exit_status = 0

        if pid in (0, -1):
            procs = self.myrrh_syscall.getallproc()
        else:
            procs = (self.myrrh_syscall.getproc(pid),)

        for p in procs:
            if options is not None:
                self.myrrh_os.shell.signal(p.pid, options)

            exit_status = self.myrrh_syscall.waitobj(p)

        if pid == -1:
            pid = 0

        return (pid, exit_status)

    def waitstatus_to_exitcode(self, status):
        if not isinstance(status, int):
            raise TypeError("an integer is required")
        return status >> 8

    def spawnv(self, mode, file, args):
        return self._spawnve(mode, file, args, env=None)

    def spawnve(self, mode, file, args, env):
        return self._spawnve(mode, file, args, env)

    def spawnl(self, mode, file, *args):
        return self._spawnve(mode, file, args, env=None)

    def spawnle(self, mode, file, *args):
        env = args[-1]
        return self._spawnve(mode, file, args[:-1], env)

    def kill(self, pid, sig):
        try:
            self.myrrh_os.shell.signal(pid, sig)

        except (OSError, IOError):
            raise
        except Exception as e:
            MOsError(self, strerror=str(e)).raised()

    class _wrap_close(metaclass=ImplPropertyClass):
        def __init__(self, stream, proc):
            self._stream = stream
            self._proc = proc

        def close(self):
            self._stream.close()
            returncode = self._proc.wait()
            if returncode == 0:
                return None
            if self.__rself__.cfg.system.os == "nt":
                return returncode
            else:
                return returncode << 8

        def __enter__(self):
            return self

        def __exit__(self, *args):
            self.close()

        def __getattr__(self, name):
            return getattr(self._stream, name)

        def __iter__(self):
            return iter(self._stream)

    def system(self, command):
        shell = list(self.getdefaultshellb(command))

        rval = self._spawnve(self.P_INTERACT, shell[0], shell, env=None)
        return rval



import typing

from myrrh.core.interfaces import abstractmethod, ABC
from myrrh.core.services.system import AbcRuntimeDelegate

from . import mbuiltins

__mlib__ = "AbcTarFile"


class _interface(ABC):
    import tarfile as local_tarfile

    @property
    @abstractmethod
    def TarFile(self) -> local_tarfile.TarFile:
        ...

    @property
    @abstractmethod
    def TarInfo(self) -> local_tarfile.TarInfo:
        ...

    @property
    @abstractmethod
    def TarError(self) -> local_tarfile.TarError:
        ...

    @property
    @abstractmethod
    def ReadError(self) -> local_tarfile.ReadError:
        ...

    @property
    @abstractmethod
    def CompressionError(self) -> local_tarfile.CompressionError:
        ...

    @property
    @abstractmethod
    def StreamError(self) -> local_tarfile.StreamError:
        ...

    @property
    @abstractmethod
    def ExtractError(self) -> local_tarfile.ExtractError:
        ...

    @property
    @abstractmethod
    def HeaderError(self) -> local_tarfile.HeaderError:
        ...

    @property
    @abstractmethod
    def ENCODING(self):
        return self.local_tarfile.ENCODING

    @property
    @abstractmethod
    def USTAR_FORMAT(self):
        return self.local_tarfile.USTAR_FORMAT

    @property
    @abstractmethod
    def GNU_FORMAT(self):
        return self.local_tarfile.GNU_FORMAT

    @property
    @abstractmethod
    def PAX_FORMAT(self):
        return self.local_tarfile.PAX_FORMAT

    @property
    @abstractmethod
    def DEFAULT_FORMAT(self):
        return self.local_tarfile.DEFAULT_FORMAT

    @property
    @abstractmethod
    def RECORDSIZE(self):
        return self.local_tarfile.RECORDSIZE

    @property
    @abstractmethod
    def GNUTYPE_LONGNAME(self):
        return self.local_tarfile.GNUTYPE_LONGNAME

    @property
    @abstractmethod
    def XHDTYPE(self):
        return self.local_tarfile.XHDTYPE

    @property
    @abstractmethod
    def BLOCKSIZE(self):
        return self.local_tarfile.BLOCKSIZE

    @property
    @abstractmethod
    def DIRTYPE(self):
        return self.local_tarfile.DIRTYPE

    @property
    @abstractmethod
    def BLKTYPE(self):
        return self.local_tarfile.BLKTYPE

    @property
    @abstractmethod
    def REGTYPE(self):
        return self.local_tarfile.REGTYPE

    @property
    @abstractmethod
    def PAX_NUMBER_FIELDS(self):
        return self.local_tarfile.PAX_NUMBER_FIELDS

    @property
    @abstractmethod
    def LENGTH_NAME(self):
        return self.local_tarfile.LENGTH_NAME

    @property
    @abstractmethod
    def LNKTYPE(self):
        return self.local_tarfile.LNKTYPE

    @property
    @abstractmethod
    def LENGTH_LINK(self):
        return self.local_tarfile.LENGTH_LINK

    @abstractmethod
    def open(self, name=None, mode="r", fileobj=None, bufsize=RECORDSIZE, **kwargs) -> local_tarfile.TarFile:
        ...

    @abstractmethod
    def is_tarfile(self, name) -> bool:
        ...

    @abstractmethod
    def stn(s, length, encoding, errors) -> typing.Any:
        ...

    @abstractmethod
    def nts(s, encoding, errors) -> typing.Any:
        ...

    @abstractmethod
    def nti(s) -> typing.Any:
        ...

    @abstractmethod
    def itn(n, digits=8, format=DEFAULT_FORMAT) -> typing.Any:
        ...

    @abstractmethod
    def calc_chksums(buf) -> typing.Any:
        ...


class AbcTarFile(_interface, AbcRuntimeDelegate):
    __frameworkpath__ = "mpython.mtarfile"

    __all__ = _interface.local_tarfile.__all__

    __delegated__ = {_interface: _interface.local_tarfile}
    __delegate_check_type__ = False

    __name__ = _interface.local_tarfile.__name__

    def __init__(self, *a, **kwa):
        mod = mbuiltins.wrap_module(self.local_tarfile, self)

        self.__delegate__(_interface, mod)



# -*- coding: utf-8 -*-
import bmy
import unittest
import io as localio

import os

from myrrh.utils.myrrh_test_init import *  # noqa: F403

setUp(f"bmy.{os.path.basename(__file__)}")  # noqa: F405

with bmy.select():
    from mlib.py import os
    from mlib.py.test.support import os_helper


def read_dist(path, entity):
    with localio.BytesIO() as buf:
        with bmy.select(entity):
            from mlib.py import os

        path = path.replace("\\", os.sep).replace("/", os.sep)
        entity.runtime.myrrh_syscall.stream_in(os.fsencode(path), buf)
        return buf.getbuffer().tobytes()


class SetupDirEntity(unittest.TestCase):
    def setUp(self):
        super().setUp()

        join = os.path.join
        self.addCleanup(os_helper.rmtree, os_helper.TESTFN)

        # Build:
        #     TESTFN/
        #       TEST1/              a file kid and two directory kids
        #         tmp1
        #         SUB1/             a file kid and a directory kid
        #           tmp2
        #           SUB11/          no kids
        #         SUB2/             a file kid and a directory kid
        #           tmp3
        #           SUB21/
        #             tmp5
        #       TEST2/
        #         tmp4              a lone file
        self.walk_path = join(os_helper.TESTFN, "TEST1")
        self.sub1_path = join(self.walk_path, "SUB1")
        self.sub11_path = join(self.sub1_path, "SUB11")
        sub2_path = join(self.walk_path, "SUB2")
        sub21_path = join(sub2_path, "SUB 21")
        tmp1_path = join(self.walk_path, "tmp1")
        tmp2_path = join(self.sub1_path, "tmp2")
        tmp3_path = join(sub2_path, "tmp3")
        tmp5_path = join(sub21_path, "tmp5")
        self.link_path = join(sub2_path, "link")
        t2_path = join(os_helper.TESTFN, "TEST2")
        tmp4_path = join(os_helper.TESTFN, "TEST2", "tmp4")
        tmp6_path = join(os_helper.TESTFN, "TEST2", "tmp6")

        self.dirs = [
            self.walk_path,
            self.sub1_path,
            self.sub11_path,
            sub2_path,
            sub21_path,
            t2_path,
        ]

        # Create stuff.
        for d in self.dirs:
            os.makedirs(d)

        self.files = [tmp1_path, tmp2_path, tmp3_path, tmp4_path, tmp5_path]
        self.filecontents = {}
        for path in self.files:
            data = b"You've got the " + path.encode() + b"file. This file is autogenerated by the myrrh project.\n"
            self.filecontents[os.path.basename(path)] = {
                "data": data,
                "size": len(data),
            }

        data = b"\0" * 5000000 + b"You've got the " + tmp6_path.encode() + b"file. This file is autogenerated by the myrrh project.\n"

        self.filecontents[os.path.basename(tmp6_path)] = {
            "data": data,
            "size": len(data),
        }
        self.files.append(tmp6_path)

        for path in self.files:
            bmy.entity().runtime.myrrh_syscall.stream_out(
                os.fsencode(path),
                localio.BytesIO(self.filecontents[os.path.basename(path)]["data"]),
            )

    def assertFiles(self, files, data):
        for d, file in zip(data, files):
            self.assertEqual(d, self.filecontents[file]["data"], "file %s not match" % file)

    def assertSizes(self, files, sizes):
        for file, size in zip(files, sizes):
            self.assertEqual(size, self.filecontents[os.path.basename(file)]["size"])

    def assertDirs(self, dirs):
        for d in self.dirs:
            self.assertIn(d[len(os_helper.TESTFN) + 1 :], dirs)


class TestCp(SetupDirEntity):
    def test_cp_dest_is_dir(self):
        for d in self.dirs:
            dest_dir = bmy.joinpath(d, "cp")
            bmy.mkdir(dest_dir)

            for f in self.files:
                bmy.cp(f, dest_dir)
                stat_src = bmy.fstat(f)
                stat_dest = bmy.fstat(bmy.joinpath(dest_dir, bmy.basename(f)))
                self.assertEqual(stat_src["size"], stat_dest["size"])

    def test_cp_dest_is_new_name(self):
        for d in self.dirs:
            dest_dir = bmy.joinpath(d, "cp")
            bmy.mkdir(dest_dir)

            for f in self.files:
                dest_file = bmy.joinpath(dest_dir, bmy.basename(f) + "_cp")
                bmy.cp(f, dest_file)
                stat_src = bmy.fstat(f)
                stat_dest = bmy.fstat(dest_file)
                self.assertEqual(stat_src["size"], stat_dest["size"])

    def test_cp_dest_is_cwd(self):
        for d in self.dirs:
            dest_dir = bmy.joinpath(d, "cp")
            bmy.mkdir(dest_dir)

            for f in self.files:
                cwd = bmy.pwd()
                dest_dir = bmy.joinpath(cwd, d, "cp")
                bmy.mkdir(dest_dir)
                try:
                    f = bmy.abspath(f)
                    bmy.chdir(dest_dir)
                    bmy.cp(f)
                    stat_dest = bmy.fstat(f)
                    stat_src = bmy.fstat(bmy.joinpath(cwd, f))
                    self.assertEqual(stat_src["size"], stat_dest["size"])
                finally:
                    bmy.chdir(cwd)

    def test_cp_n_sources(self):
        cwd = bmy.pwd()
        for d in self.dirs:
            dest_dir = bmy.joinpath(cwd, d, "cp2")
            bmy.mkdir(dest_dir)

            bmy.cp(self.files, dest_dir)
            for f in self.files:
                stat_src = bmy.fstat(f)
                stat_dest = bmy.fstat(bmy.joinpath(dest_dir, bmy.basename(f)))
                self.assertEqual(stat_src["size"], stat_dest["size"])


class TestFs(SetupDirEntity):
    # 'edit', 'get', 'push', 'transfer',
    def test_chdir_pwd(self):
        cwd = bmy.pwd()
        for d in self.dirs:
            try:
                bmy.chdir(d)
                self.assertEqual(bmy.pwd(), bmy.joinpath(cwd, d))
            finally:
                bmy.chdir(cwd)

        for d in (bmy.abspath(d) for d in self.dirs):
            try:
                bmy.chdir(d)
                self.assertEqual(bmy.pwd(), bmy.joinpath(cwd, d))
            finally:
                bmy.chdir(cwd)

    def test_fstat(self):
        for f in self.files:
            bmy.fstat(f)

        for d in self.dirs:
            bmy.fstat(d)

    def test_lsdir(self):
        dirs = bmy.lsdir(os_helper.TESTFN)

        self.assertListEqual(dirs, ["TEST1", "TEST2"])

        self.assertRaises(NotADirectoryError, bmy.lsdir, self.files[0])

        dirs2 = bmy.lsdir(bmy.joinpath(self.dirs[0], "t*"))

        self.assertEqual(len(dirs2), 1)
        self.assertEqual(dirs2[0], bmy.joinpath(os_helper.TESTFN, "TEST1", "tmp1"))

    def test_mkdir(self):
        for d in self.dirs:
            path = bmy.joinpath(self.dirs[0], d)
            bmy.mkdir(path)
            stat = bmy.fstat(path)
            self.assertEqual(stat["file"], path)
            self.assertTrue(stat["access"].startswith("d"))


class TestRm(SetupDirEntity):
    def test_rm_file(self):
        for f in self.files:
            bmy.rm(f)
            self.assertRaises(FileNotFoundError, bmy.fstat, f)

    def test_rm_files(self):
        bmy.rm(self.files)
        for f in self.files:
            self.assertRaises(FileNotFoundError, bmy.fstat, f)

    def test_rm_dirs(self):
        bmy.rm(self.dirs)
        for d in self.dirs:
            self.assertRaises(FileNotFoundError, bmy.fstat, d)


class TestMove(SetupDirEntity):
    def test_move(self):
        path = bmy.joinpath(self.dirs[0], "moved")
        bmy.mkdir(path)
        for f in self.files:
            new_path = bmy.joinpath(path, bmy.basename(f) + "_moved")
            bmy.move(f, new_path)
            stat = bmy.fstat(new_path)
            self.assertEqual(stat["file"], new_path)

    def test_move_dest_is_dir_one_file(self):
        path = bmy.joinpath(self.dirs[0], "moved")
        bmy.mkdir(path)

        for f in self.files:
            new_path = bmy.joinpath(path, bmy.basename(f))
            bmy.move(f, path)
            stat = bmy.fstat(new_path)
            self.assertEqual(stat["file"], new_path)

    def test_move_dest_is_dir_many_files(self):
        path = bmy.joinpath(self.dirs[0], "moved")
        bmy.mkdir(path)
        bmy.move(self.files, path)

        for f in self.files:
            new_path = bmy.joinpath(path, bmy.basename(f))
            stat = bmy.fstat(new_path)
            self.assertEqual(stat["file"], new_path)


class TestReadWrite(SetupDirEntity):
    def test_read(self):
        data = []
        for f in self.files:
            data.append(bmy.read(f, binary=True))

        self.assertFiles((bmy.basename(f) for f in self.files), data)

    def test_write(self):
        for f, d in self.filecontents.items():
            bmy.write(bmy.joinpath(self.dirs[0], f), d["data"])

        for f, d in self.filecontents.items():
            data = bmy.read(bmy.joinpath(self.dirs[0], f), binary=True)
            self.assertEqual(data, d["data"])


if __name__ == "__main__":
    unittest.main()



import posix

from myrrh.framework.mpython._mosprocess import AbcOsProcess

__mlib__ = "OsProcess"


class OsProcess(AbcOsProcess):
    SIGKILL = 9
    WIFSTOPPED = posix.WIFSTOPPED
    WSTOPSIG = posix.WSTOPSIG
    WNOHANG = posix.WNOHANG



import collections
import inspect
import functools

from concurrent.futures import ThreadPoolExecutor, wait

from . import cfg_init


__all__ = (
    "myrrh_group_iter",
    "myrrh_group_keys",
    "MyrrhGroup",
    "is_myrrh_group",
    "myrrh_group",
    "myrrh_group_sync",
    "myrrh_group_async",
    "myrrh_group_async_member",
)


def myrrh_group_iter(iterable):
    return iter(getattr(iterable, "_t_", iterable))


def myrrh_group_keys(iterable):
    return tuple(getattr(iterable, "_f_", iterable))


def myrrh_group_values(iterable):
    return tuple(getattr(iterable, "_t_", iterable))


def is_myrrh_group(group):
    return isinstance(group, MyrrhGroup)


def myrrh_group(func, keys=None):
    return functools.wraps(func)(MyrrhGroup(func, keys=keys))


def myrrh_group_sync(func, keys=None):
    return functools.wraps(func)(MyrrhGroup(func, keys=keys)._s_)


def myrrh_group_async(func, keys=None):
    return functools.wraps(func)(MyrrhGroup(func, keys=keys)._as_)


def myrrh_group_async_member(method):
    method.async_ = True
    return method


def myrrh_group_sync_member(method):
    method.async_ = False
    return method


class MyrrhGroupPropertySync:
    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return MyrrhGroup(
            *(getattr(instance._d_[k], self.name) for k in instance._f_),
            keys=instance._f_,
        )

    def __set__(self, instance, value):
        for k in instance._f_:
            setattr(instance._d_[k], self.name, value)


class MyrrhGroupPropertyASync:
    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        with MyrrhGroupPool(instance) as pool:
            return MyrrhGroup(
                *pool.map(lambda k: getattr(instance._d_[k], self.name), instance._f_),
                keys=instance._f_,
            )

    def __set__(self, instance, value):
        with MyrrhGroupPool(instance) as pool:
            pool.map(lambda k: setattr(instance._d_[k], self.name, value), instance._f_)


class MyrrhGroupMeta(type):
    __async__ = cfg_init("group_threaded", False, section="myrrh.core.services.system")
    __max_workers__ = cfg_init("group_max_concurrent_threads", 5, section="myrrh.core.services.system")

    def __prepare__(name, bases, *, namedtuple_, async_=None):
        _dict = dict()
        if async_ is None:
            async_ = MyrrhGroupMeta.__async__

        _dict["__namedtuple__"] = namedtuple_()
        _dict["__async__"] = async_
        _dict["__max_workers__"] = MyrrhGroupMeta.__max_workers__

        if namedtuple_._field_defaults:
            master_item = namedtuple_()[0]

            for k, m in inspect.getmembers_static(master_item):
                if not k.startswith("_"):
                    _dict[k] = (MyrrhGroupPropertyASync if getattr(m, "async_", False) else MyrrhGroupPropertySync)()

            if callable(master_item):
                _dict["__call__"] = MyrrhGroupMeta.__async_group_call__ if getattr(master_item, "async_", async_) else MyrrhGroupMeta.__sync_group_call__

            try:
                iter(master_item)
                _dict["__iter__"] = MyrrhGroupMeta.__group_iter__
            except TypeError:
                ...

            for spe_attr in (
                "__getitem__",
                "__setitem__",
                "__delitem__",
                "__missing__",
                "__reversed__",
                "__contains__",
                "__enter__",
                "__exit__",
            ):
                if hasattr(master_item, spe_attr):
                    _dict[spe_attr] = (MyrrhGroupPropertyASync if getattr(m, "async_", False) else MyrrhGroupPropertySync)()

            for oper in (
                "__add__",
                "__sub__",
                "__mul__",
                "__truediv__",
                "__floordiv__",
                "__mod__",
                "__divmod__",
                "__pow__",
                "__lshift__",
                "__rshift__",
                "__and__",
                "__xor__",
                "__or__",
                "__radd__",
                "__rsub__",
                "__rmul__",
                "__rtruediv__",
                "__rfloordiv__",
                "__rmod__",
                "__rdivmod__",
                "__rpow__",
                "__rlshift__",
                "__rrshift__",
                "__rand__",
                "__rxor__",
                "__ror__",
                "__iadd__",
                "__isub__",
                "__imul__",
                "__itruediv__",
                "__ifloordiv__",
                "__imod__",
                "__idivmod__",
                "__ipow__",
                "__ilshift__",
                "__irshift__",
                "__iand__",
                "__ixor__",
                "__ior__",
                "__neg__",
                "__pos__",
                "__abs__",
                "__invert__",
                "__match_args__",
            ):
                if hasattr(master_item, oper):
                    _dict[oper] = (MyrrhGroupPropertyASync if getattr(m, "async_", False) else MyrrhGroupPropertySync)()
            """
            for unsupported in ('__complex__', '__int__', '__float__', '__index__',
                           '__round__', '__trunc__', '__floor__', '__ceil__',
                           '__length_hint__', '__len__):
                if hasattr(master_item, unsupported):
                    _dict[unsupported] =  (MyrrhGroupPropertyASync if getattr(m, 'async_', False) else MyrrhGroupPropertySync)()
            """

            _dict["__item_master__"] = property(fget=lambda cls: cls.__namedtuple__[0])
            _dict["__item_class__"] = getattr(master_item, "__item_class__", None) or getattr(master_item, "__class__") or None.__class__

        else:
            _dict["__item_master__"] = None
            _dict["__item_class__"] = None.__class__

        return _dict

    def __group_iter__(self):
        iter_ = myrrh_group_sync(iter)(self)
        try:
            while True:
                yield myrrh_group_sync(next)(iter_)

        except StopIteration:
            ...

    def __sync_group_call__(self, *a, **kwa):
        keys, vals, args, kwargs = MyrrhGroupMeta._group_members(self, a, kwa)
        futures = [(v, a, kwa) for v, a, kwa in zip(vals, args, kwargs)]

        results = list()
        rsult_keys = list()
        keys = list(reversed(keys))

        for v, a, kwa in futures:
            try:
                results.append(v(*a, **kwa))
                rsult_keys.append(keys.pop())

            except StopIteration:
                keys.pop()

        if len(rsult_keys) == 0:
            raise StopIteration

        return MyrrhGroup(*results, keys=rsult_keys)

    def __async_group_call__(self, *a, **kwa):
        keys, vals, args, kwargs = MyrrhGroupMeta._group_members(self, a, kwa)

        with MyrrhGroupPool(keys) as pool:
            futures = [pool.submit(v, *a, **kwa) for v, a, kwa in zip(vals, args, kwargs)]

        wait(futures)

        results = list()
        rsult_keys = list()
        keys = list(reversed(keys))

        for f in futures:
            try:
                results.append(f.result())
                rsult_keys.append(keys.pop())
            except StopIteration:
                keys.pop()

        if len(rsult_keys) == 0:
            raise StopIteration

        return MyrrhGroup(*results, keys=rsult_keys)

    def _group_members(self, args, kwargs):
        vals = []
        group_args = []
        group_kwargs = []

        for group in filter(lambda a: isinstance(a, MyrrhGroup), args):
            keys = group._f_
            break
        else:
            for group in filter(lambda a: isinstance(a, MyrrhGroup), kwargs.values()):
                keys = group._f_
                break
            else:
                keys = self._f_

        try:
            for k in keys:
                val = self.__namedtuple__._field_defaults.get(k) or self.__item_master__
                vals.append(val)
                group_args.append([(a if not isinstance(a, MyrrhGroup) else a._d_[k]) for a in args])
                group_kwargs.append({ka: (va if not isinstance(va, MyrrhGroup) else va._d_[k]) for ka, va in kwargs.items()})

            return keys, vals, group_args, group_kwargs

        except KeyError:
            pass

        raise ValueError("multiple MyrrhRuntimeGroup in arguments but keys are not the same")


class MyrrhGroup:
    def __new__(cls, *items, keys=None, async_=None):
        keys = keys or tuple(map(lambda i: f"f{i}", range(0, items and len(items) or 1)))

        if not items:
            items = keys

        keys = getattr(keys, "_f_", keys)
        items = myrrh_group_values(items)

        NamedTuple = collections.namedtuple("MyrrhRuntimeGroup", keys, defaults=items)

        class _MyrrhRuntimeGroup(MyrrhGroup, namedtuple_=NamedTuple, async_=async_, metaclass=MyrrhGroupMeta):
            def __new__(cls, *items, keys=None, async_=None):
                return object.__new__(cls)

        return _MyrrhRuntimeGroup()

    def __init_subclass__(cls, namedtuple_, async_):
        ...

    @property  # type: ignore[misc]
    def __class__(self):
        return self.__item_class__

    @property
    def _f_(self):
        return self.__namedtuple__._fields

    @property
    def _d_(self):
        return self.__namedtuple__._asdict()

    @property
    def _t_(self):
        return tuple(self.__namedtuple__)

    @property
    def _as_(self, *, max_workers=None):
        group = MyrrhGroup(*self._t_, keys=self._f_, async_=True)
        if max_workers:
            group.__max_workers__ = max_workers

        return group

    @property
    def _s_(self):
        return MyrrhGroup(*self._t_, keys=self._f_, async_=False)


class MyrrhGroupPool(ThreadPoolExecutor):
    def __init__(self, keys):
        if not len(keys):
            raise RuntimeError("invalid keys")
        super().__init__(len(keys), f"Myrrh{keys}")



import re
from functools import cached_property

from myrrh.utils import mshlex
from myrrh.utils import merrno as errno

from myrrh.core.services.system import AbcMyrrhOs, ExecutionFailureCauseRVal


class MyrrhOs(AbcMyrrhOs):
    _curdirb_ = b"."
    _pardirb_ = b".."
    _extsepb_ = b"."
    _sepb_ = b"/"
    _pathsepb_ = b":"
    _defpathb_ = b":/bin:/usr/bin"
    _altsepb_ = None
    _devnullb_ = b"/dev/null"
    _linesepb_ = b"\n"

    import posixpath

    isabs = staticmethod(posixpath.isabs)  # type: ignore[assignment]
    normpath = staticmethod(posixpath.normpath)  # type: ignore[assignment]
    joinpath = staticmethod(posixpath.join)  # type: ignore[assignment]
    basename = staticmethod(posixpath.basename)  # type: ignore[assignment]
    dirname = staticmethod(posixpath.dirname)  # type: ignore[assignment]

    _bin_dict: dict[bytes, bytes] | None = None
    _std_dict = {
        b"find": b"/usr/bin/find",
        b"ln": b"/bin/ln",
        b"chown": b"/bin/chown",
        b"chgrp": b"/bin/chgrp",
        b"chmod": b"/bin/chmod",
        b"mv": b"/bin/mv",
        b"realpath": b"/usr/bin/realpath",
        b"touch": b"/usr/bin/touch",
        b"cat": b"/bin/cat",
        b"pwd": b"/bin/pwd",
        b"sh": b"/bin/sh",
        b"set": b"set",
        b"which": b"/usr/bin/which",
        b"dirname": b"/usr/bin/dirname",
        b"id": b"/usr/bin/id",
        b"stat": b"/usr/bin/stat",
        b"echo": b"/bin/echo",
        b"truncate": b"/usr/bin/truncate",
        b"sleep": b"/bin/sleep",
        b"ps": b"/bin/ps",
        b"xargs": b"/usr/bin/xargs",
        b"cp": b"/bin/cp",
        b"tar": b"/bin/tar",
        b"mkdir": b"/bin/mkdir",
        b"rm": b"/bin/rm",
    }

    def _getbinb_(self):
        self._bin_dict = dict(self._std_dict)
        out, _, rval = self._delegate_.shell.execute(b"for p in /bin/* /usr/bin/*; do echo $p; done")
        if not rval:
            bin = {self.basename(b): b for b in out.split()}
            self._bin_dict.update(bin)

        return self._bin_dict

    def _getdefaultshellb_(self):
        return self.getenvb().get(b"SHELL", self._getbinb_()[b"sh"])

    def _getdefaultshellargsb_(self):
        return (b"-c",)

    def formatshellargs(self, args, *, defaultargs=None):
        if isinstance(args, str):
            args = args.encode()
        return (
            *defaultargs,
            mshlex.list2cmdlineb(args).strip(b'"') if isinstance(args, (list, tuple)) else args,
        )

    def _getdefaultencoding_(self):
        return "utf-8"

    def _fsencoding_(self):
        return "utf-8"

    def _fsencodeerrors_(self):
        return "surrogateescape"

    def sh_escape_bytes(self, string):
        return mshlex.quote(string)
        # return '%s' % string.replace('"', r'\"')

    def _localecode_(self):
        out, err, rval = self.cmd(b"%(echo)s $LANG")
        ExecutionFailureCauseRVal(self, err, rval, 0, errno=errno.EFAULT).check()  # avoid translate error

        return out.split(".")[0]

    def _getcwdb_(self):
        out, err, rval = self.cmdb(b"%(pwd)s")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        return out

    def _gettmpdirb_(self):
        out, err, rval = self.cmdb(b'%(echo)s "$TMPDIR:$TEMP:$TMP:`[ -d /tmp ] && %(echo)s /tmp`:`[ -d /var/tmp ] && %(echo)s /var/tmp`:`[ -d /usr/tmp ] && %(echo)s /usr/tmp`"')
        ExecutionFailureCauseRVal(self, err, rval, 0).check()

        dirlist = [dir for dir in filter(None, out.split(b":"))]
        return dirlist[0] if dirlist else self._getcwdb_()

    def _getreadonlyenvb_(self):
        return (b"PPID", b"SHELLOPTS")

    def _envb_(self):
        out, err, rval = self._delegate_.shell.execute(b"set")
        ExecutionFailureCauseRVal(self, err, rval, 0).check()
        env = {k: v for k, v in re.findall(b"(?P<k>[^\\n=]*)=(?P<v>'[^']*'\\n|.*\\n)", out)}
        env = {k: v.rstrip(b"\n") for k, v in env.items()}
        env = {k: v[1:-1] if len(v) > 2 and v[:1] == b"'" and v[-1:] == b"'" else v for k, v in env.items()}  # PIF! PIF patch
        return env

    @cached_property
    def _errno_localized_mapping(self):
        return errno.errno_create_localized_mapping(self.getdefaultlocale()[0])

    def default_errno_from_msg(self, err):
        err = self.shencode(err)
        return errno.errno_from_msgb(err, map=self._errno_localized_mapping)

    environkeyformat = None  # type: ignore[assignment]
    error_translate = default_errno_from_msg



import pprint
import click

__all__ = ["secret", "password"]


class _F:
    from myrrh.utils.secrets import set_key, clean_all_keys, clean_key_history, dump, backup_key, delete_key, use_key, rename_key, encrypt, decrypt  # type: ignore[misc]

    @classmethod
    def new_password(cls):
        click.password_option


def pecho(s, *a, **kwa):
    click.echo(pprint.pformat(s), *a, **kwa)


@click.option("-c", "--clean-history", "clean", flag_value="clean_key_history", default=False)
@click.option("--clean-all", "clean", flag_value="clean_all_keys", default=False)
@click.option("-d", "--dump", "dump", is_flag=True, default=False)
@click.option("-g", "--gen-key", "gen", is_flag=True, default=False)
@click.option("-f", "--force", is_flag=True, default=False)
@click.option("-b", "--backup-key", type=str, default=None)
@click.option("--delete-key", type=str, default=None)
@click.option("--rename-key", nargs=2, type=str, default=None)
@click.option("-u", "--use-key", type=str, default=None)
def secret(gen, clean, dump, backup_key, force, delete_key, use_key, rename_key):
    if clean:
        getattr(_F, clean)()

    if delete_key:
        _F.delete_key(delete_key)

    if rename_key:
        _F.rename_key(*rename_key)

    if backup_key:
        _F.backup_key(backup_key, overwrite=force)

    if gen:
        _F.set_key()

    if use_key:
        _F.use_key(use_key)

    if dump:
        pecho(_F / dump())


def prompt_password(ctx, param, value):
    if value == "-":
        value = click.prompt("New", hide_input=True, show_choices=False, confirmation_prompt=True)
    return value


@click.option(
    "-c",
    "encrypt",
    type=str,
    default="",
    callback=prompt_password,
    help="encrypt the given string",
)
@click.option("-d", "decrypt", type=str, default="", help="decrypt the given string")
@click.option("-k", "key", type=str, default="key")
def password(encrypt, decrypt, key):
    if encrypt:
        print(_F.encrypt(encrypt, key))
    if decrypt:
        print(_F.decrypt(decrypt, key))


if __name__ == "__main__":

    @click.group
    def main():
        ...

    main.command(secret)
    main.command(hash)

    main()



# -*- coding: utf-8 -*-
"""
**Myrrh password manager module**

    This module contains necessary functionalities for **volatile** storing of passwords.
"""

import getpass

__all__ = ["Passwd"]


class CacheData(object):
    def __init__(self):
        pass


class Passwd:
    __hostlist: dict[str, dict[str, str]] = {}

    class NotAcquired(Exception):
        @classmethod
        def raised(cls):
            raise cls()

    @classmethod
    def getPass(cls, host, user, pwd=None):
        if host not in Passwd.__hostlist:
            Passwd.__hostlist[host] = {}

        if user in Passwd.__hostlist[host]:
            return Passwd.__hostlist[host][user]

        pwd = pwd if pwd is not None else getpass.getpass(("Please enter needed password for user %s on %s\npassword:" % (user, host)).encode("ascii"))

        Passwd.__hostlist[host][user] = ("%s" % pwd).strip()

        return Passwd.__hostlist[host][user]

    @classmethod
    def getData(cls, host, name, value=None, acquire=lambda: Passwd.NotAcquired.raised()):
        if host not in Passwd.__hostlist:
            Passwd.__hostlist[host] = {}

        if value is None and name in Passwd.__hostlist[host]:
            return Passwd.__hostlist[host][name]

        if name in Passwd.__hostlist[host] and value != Passwd.__hostlist[host][name]:
            Passwd.__hostlist[host][name] = value

        try:
            Passwd.__hostlist[host][name] = acquire() if value is None else value
        except Passwd.NotAcquired:
            return None

        return Passwd.__hostlist[host][name]

    @classmethod
    def invalidPass(cls, host, user=None):
        if user is None:
            Passwd.__hostlist.pop(host)
        else:
            Passwd.__hostlist[host].pop(user, None)



from __future__ import annotations

import typing
import datetime

import pydantic
import pydantic.json

from myrrh.core.services import cfg_init

__all__ = ["NoneItem", "BaseItem", "GenericItem", "NoneItemType"]

ItemT = typing.TypeVar("ItemT")


class BaseItem(pydantic.BaseModel, typing.Generic[ItemT]):
    _utc = cfg_init("dom_in_utc", False, section="myrrh.item")

    type_: ItemT = None  # type: ignore[assignment]

    DOM: pydantic.types.PastDatetime | None = None
    SLL: datetime.timedelta | None = None
    UBD: datetime.datetime | None = None
    UTC: bool = _utc

    model_config = pydantic.ConfigDict(extra="forbid", validate_assignment=True, frozen=True)

    def now(self):
        return datetime.datetime.utcnow if self.UTC else datetime.datetime.now

    def __bool__(self):
        return self.UBD is None or self.UBD >= self.now()

    @classmethod
    def _calc_ubd(cls, dom, sll):
        if sll == datetime.timedelta():
            return dom

        try:
            return dom + sll
        except OverflowError:
            pass

        return None

    @classmethod
    def _calc_dom(cls, dom, now) -> datetime.datetime:
        return dom if dom and dom < now else now

    @classmethod
    def _type_(cls):
        annotation = typing.get_args(cls.model_fields["type_"].annotation)

        if annotation:
            return annotation[0]

        return "generic"

    @pydantic.model_validator(mode="before")
    @classmethod
    def model_validation_before(cls, data):
        type_ = data.get("type_") or cls._type_()
        data["type_"] = type_

        utc = data.get("UTC", cls._utc)

        SLL = data.get("SLL")
        if SLL is not None and SLL < datetime.timedelta():
            raise ValueError("SLL must be a positive value")

        DOM = data.get("DOM")
        if SLL and DOM is None:
            now = datetime.datetime.utcnow() if utc else datetime.datetime.utcnow()
            DOM = cls._calc_dom(DOM, now)
            data["DOM"] = DOM

        UBD = data.get("UBD")
        if SLL and UBD is None:
            UBD = cls._calc_ubd(DOM, SLL)
            if UBD is not None:
                data["UBD"] = UBD

        return data

    @staticmethod
    def _serialize(*a, **kwa):
        pass

    @staticmethod
    def _val(val, **kwa):
        return val

    def model_dump_json(self, *, by_alias=False, exclude_defaults=True, **dumps_kwargs):
        return super().model_dump_json(by_alias=by_alias, exclude_defaults=exclude_defaults, **dumps_kwargs)

    def __hash__(self):
        return hash(self.type_)

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, item, value):
        setattr(self, item, value)

    def __repr__(self):
        return self.model_dump_json()


class GenericItem(BaseItem[str]):
    model_config = pydantic.ConfigDict(extra="allow")

    type_: str

    @pydantic.field_validator("type_")
    @classmethod
    def type_validate(cls, type_):
        from .registry import ItemRegistry

        if type_ in ItemRegistry.__warehouse_items__:
            raise ValueError(f'"{type_}" is not a valid item')

        return type_


class NoneItemType(BaseItem[typing.Literal["null"]]):
    def __bool__(self):
        return False

    def __getattr__(self, item):
        return self

    def __str__(self):
        return "na"


NoneItem = NoneItemType(type_="null")



import colorama
import logging
import os
import signal

import prompt_toolkit

import bmy

from myrrh.utils import mstring
from myrrh.core.services import log, myrrh_sys, cfg_get, cfg_set, cfg_del
from myrrh.core.services import rebase as rebase_

from myrrh.tools.myrrhc_ext import myrrhc_cmds, cmd
from myrrh.tools.myrrhc import completion
from myrrh.tools.myrrhc import types

from prompt_toolkit.application.run_in_terminal import run_in_terminal

colorama.init()

cmd.option_cfg_init("editor")


@myrrhc_cmds.command(hidden=True, cls=cmd.DefaultCommand, group=myrrhc_cmds)
@cmd.pass_context
def _default(ctx):
    if ctx.cmdline:
        if not bmy.current():

            def _in_terminal():
                os.system(ctx.cmdline)

            run_in_terminal(_in_terminal)
        else:
            system([ctx.cmdline], standalone_mode=False, parent=ctx)


@myrrhc_cmds.command()
@cmd.argument("level", shell_complete=["debug", "info", "error", "warning", "disable"])
def debug(level):
    "(De)Activate Log at LEVEL"

    bmy.debug(level=None if level == "disable" else level)
    cmd.info("debug mode set to %s" % ("disabled" if log.disabled else logging.getLevelName(log.level)))


@myrrhc_cmds.command()
@cmd.argument(
    "path",
    default=None,
    type=cmd.Path(),
    required=False,
    shell_complete=completion.local_path_completer(file_filter=lambda f: f.is_dir() or f.name == myrrh_sys),
)
def rebase(path):
    "Rebase myrrh system variables on file or directory where myrrh.msys is present"
    import json

    cmd.info(json.dumps(rebase_(path), indent=2))


def _setcfg_completer_key(ctx, cli, incomplet):
    section = ctx.params["section"]

    return [k for k in cfg_get(section=section, default={}) if not k.startswith("@")]


def _setcfg_completer_section(ctx, cli, incomplet):
    return [k.strip("__") for k in cfg_get() if k.startswith("__")]


def _setcfg_completer_value(ctx, cli, incomplet):
    section = ctx.params["section"]
    key = ctx.params["key"]
    value = cfg_get(key, section=section) if key else None
    return [value] if isinstance(value, str) else []


@myrrhc_cmds.command()
@cmd.option(
    "-s",
    "--section",
    default="",
    type=types.STRING,
    required=False,
    shell_complete=_setcfg_completer_section,
)
@cmd.option("-r", "--rebase", flag_value=True)
@cmd.option("-e", "--edit", flag_value=True)
@cmd.argument(
    "key",
    default=None,
    type=types.STRING,
    required=False,
    shell_complete=_setcfg_completer_key,
)
@cmd.argument(
    "value",
    default=None,
    type=types.STRING,
    required=False,
    nargs=-1,
    shell_complete=_setcfg_completer_value,
)
def setcfg(key, edit, rebase, value, section):
    "Add or delete config entry in myrrh system variables (set no value to delete)"

    if rebase:
        rebase_()

    if not key and not edit:
        import json

        cmd.info(json.dumps(cfg_get(section=section), indent=2))
        return

    if key:
        if not value:
            cfg_del(key, section)
        else:
            cfg_set(key, mstring.str2int(" ".join(value)), section)

    if edit:
        import click

        click.edit(filename=cfg_get("@mbase@"), editor=cmd.option_cfg_get("editor"))
        rebase_()


@myrrhc_cmds.command()
@cmd.argument("path", type=cmd.Path(), shell_complete=completion.local_path_completer())
def load(path):
    "Load an entity"
    eid = bmy.load(
        path,
    )
    bmy.select(eid)
    cmd.success(f'new entity "{eid}" loaded , build required.')


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.option("--full", flag_value=False)
@cmd.argument("path", type=cmd.Path(), shell_complete=completion.local_path_completer())
def save(path, eid, full):
    "Save an entity"
    bmy.save(path, full=full, eid=eid)
    cmd.success(f'entity "{eid}" saved on path.')


@myrrhc_cmds.command()
@cmd.pass_option_eid
def build(eid):
    "Build an entity"
    eid = bmy.build(eid=eid)
    with cmd.EidOuput(eid=eid) as output:
        cmd.info("entity built", output=output)


@myrrhc_cmds.command(name="list")
def list():
    for eid in bmy.eids():
        cmd.info(cmd.format_eid(eid))


@myrrhc_cmds.command()
@cmd.argument("eids", nargs=-1, shell_complete=completion.eids_completer)
def select(eids):
    bmy.select(eid=eids)


@myrrhc_cmds.command()
def unselect():
    bmy.unselect()


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("attribute", required=False, nargs=-1)
def info(eid, attribute):
    "Print information on a entity"

    info = bmy.info(attribute or None, eid=eid)
    if info is None:
        cmd.info("no info available")
        return

    with cmd.EidOuput(eid) as output:
        for eid, val in info._d_.items():
            lenjust = len(max(val.keys(), key=len))
            for k, v in val.items():
                if v:
                    info_str = v.model_dump_json(indent=2) if hasattr(v, "model_dump_json") else str(v)
                else:
                    info_str = "na"
                cmd.info(f"{k.ljust(lenjust)} {info_str}", output=output, eid=eid)


@myrrhc_cmds.command()
def adv():
    "Myrrhc advanced mode"

    async def pyembed():
        import importlib

        locals = {e: bmy.entity(e) for e in bmy.eids()}
        locals.update(
            {
                "bmy": bmy,
                "log": log,
                "exit": lambda: importlib.import_module("sys").exit(),
                "quit": lambda: importlib.import_module("sys").exit(),
            }
        )

        banner = """Myrrhc Advanced mode:
    exit() to return to normal BmyMode
    <entity_id>. to access entity instance

    list of entities:
    """
        for e in bmy.eids():
            banner = "%s\t%s\n" % (banner, e)

        # iconsole = code.InteractiveConsole(locals)
        # for line in pre_src.splitlines():
        #     iconsole.push(line)
        # iconsole.interact(banner=banner)

        cmd.info(banner)
        from ptpython.repl import embed

        try:
            await embed(globals(), locals, return_asyncio_coroutine=True, patch_stdout=True)
        except EOFError:
            pass
        except SystemExit:
            pass
        except KeyboardInterrupt:
            pass

    return pyembed


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.option(
    "-w",
    "--wait",
    is_flag=True,
    default=False,
    help="wait for boot completion (default: no wait)",
)
def boot(eid, wait):
    "Boot the entity"
    bmy.boot(eid=eid, wait=wait)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.option(
    "-w",
    "--wait",
    is_flag=True,
    default=False,
    help="wait for reboot completion (default: no wait)",
)
@cmd.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="force reboot (default: do not force)",
)
def reboot(eid, wait, force):
    "Soft reboot of the entity"
    bmy.reboot(wait=wait, force=force, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.option(
    "-w",
    "--wait",
    is_flag=True,
    default=False,
    help="wait for entity halted (default: no wait)",
)
@cmd.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="force halt (default: do not force)",
)
def halt(eid, wait, force):
    "Soft halt of the entity"
    bmy.halt(eid=eid, wait=wait, force=force)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("name", type=types.STRING)
def snap(eid, name):
    "Take a snapshot of the entity"
    bmy.snap(name, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("name", type=types.STRING)
def resnap(eid, name):
    "Restore a snapshot of the entity"
    bmy.resnap(name, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("name", type=types.STRING)
def desnap(eid, name):
    "Delete a snapshot of the entity"
    bmy.desnap(name, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
def snaps(eid):
    "List available snapshots on an entity"
    names = bmy.snaps(eid=eid)

    for e, name in names._d_.items():
        cmd.info("%s:" % e)
        cmd.info("\n".join(name))


@myrrhc_cmds.command()
@cmd.pass_option_eid
def csnap(eid):
    "name of current entity snapshot"
    name = bmy.csnap(eid=eid)
    cmd.info(name)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("command", shell_complete=completion.path_executor())
def system(eid, command):
    "System execution"
    cmd.info('call system shell with "%s"' % command, bold=True)
    bmy.system(command, eid=eid)


@myrrhc_cmds.command(name="exec")
@cmd.pass_option_eid
@cmd.pass_context
@cmd.option(
    "-c",
    "--count",
    type=types.INT,
    help="stop after count execution of the command",
    default=1,
)
@cmd.option(
    "-i",
    "--interval",
    type=types.FLOAT,
    help="wait interval between execution of two commands in second",
    default=None,
)
@cmd.option(
    "-t",
    "--ttl",
    type=types.FLOAT,
    help="command execution time to live in second",
    default=None,
)
@cmd.option(
    "-w",
    "--timeout",
    type=types.FLOAT,
    help="time to wait before stop execution loop in second",
    default=None,
)
@cmd.option(
    "-d",
    "--duration",
    type=types.INT,
    help="execute command until end of duration",
    default=None,
)
@cmd.option(
    "-mc",
    "--memory-clear",
    is_flag=True,
    help="clear all entry in execution memory",
    default=False,
)
@cmd.option(
    "-mce",
    "--memory-clear-element",
    type=types.INT,
    help="remove the given PID from execution memory",
    default=None,
)
@cmd.argument("command", shell_complete=completion.path_executor())
@cmd.argument("params", nargs=-1, required=False, shell_complete=completion.path_completer())
class _exec:
    """
    Execute a command into selected entity

    Special command:
        ? [pid]: execution memory, if pid then show only data related to the selected process from execution memory
    """

    def _print_pid(self, pid):
        for lastexec in bmy.execute.mem:
            for e in lastexec.calls:
                if e.proc.pid == pid:
                    break
            else:
                continue

            col = (10, 22, 10, 8)
            cmd.print_table(('"%s"' % lastexec.cmd, "state", "time", "rval"), col, bold=True)
            cmd.print_table(
                (
                    "[%d]" % e.proc.pid,
                    "%(state)s" % e,
                    "%(time).5f" % e,
                    "%(rval)s" % e,
                ),
                col,
            )

            cmd.info("---state---", bold=True)
            cmd.info(e.state)
            cmd.info("---stdout--", bold=True)
            if e.out:
                cmd.info(e.out)
            else:
                cmd.info("<empty>")
            cmd.info("---stderr--", bold=True)
            if e.err:
                cmd.error(e.err)
            else:
                cmd.info("<empty>")

    def __init__(
        self,
        ctx,
        count,
        duration,
        interval,
        ttl,
        timeout,
        command,
        params,
        eid,
        memory_clear,
        memory_clear_element,
    ):
        self._exec = None

        if memory_clear_element is not None:
            bmy.execute_mem_ce(memory_clear_element)

        if memory_clear:
            bmy.execute_mem_c()

        if command != "?":
            cmd.info(
                'execute "%s" on %s' % (" ".join([command, *params]), cmd.format_eid(eid)),
                bold=True,
            )
            cmd.info()
            self._exec = bmy.execute(
                " ".join([command, " ".join(params)]),
                count=count,
                timeout=timeout,
                interval=interval,
                duration=duration,
                ttl=ttl,
                raiseonttl=False,
                raiseontimeout=False,
                eid=eid,
            )
            self._pids = None
        else:
            self._pids = params

    async def __call__(self):
        pids = self._pids and [mstring.str2int(p) for p in self._pids] or None

        if pids and len(pids) == 1:
            self._print_pid(pids[0])
            return

        execs = bmy.execute.mem if self._exec is None else bmy.groupvalues(self._exec)

        col = (15, 22, 10, 22, 8)
        for lastexec in execs:
            cmd.print_table(
                ('"%s"' % lastexec.cmd, "state", "time", "outputs", "rval"),
                col,
                bold=True,
            )

            count = 0
            for e in lastexec:
                if not pids or e.pid in pids:
                    count += 1
                    cmd.print_table(
                        (
                            "[%d]" % e.proc.pid,
                            "%(state)s" % e,
                            "%(time).5f" % e,
                            '"%(out)s"' % e,
                            "%(rval)s" % e,
                        ),
                        col,
                    )
                    cmd.print_table(("", "", "", '"%(err)s"' % e, ""), col)

            cmd.info(
                "%(label_ncalls)s: %(ncalls)d, %(label_times)s: %(time)0.1fs, %(label_means)s: %(means)0.5fs, %(label_var)s: %(var)0.5fs"
                % {
                    "label_ncalls": cmd.style("Total calls", bold=True),
                    "label_times": cmd.style("Total times", bold=True),
                    "label_means": cmd.style("Means", bold=True),
                    "label_var": cmd.style("Variance", bold=True),
                    "ncalls": lastexec.ncalls,
                    "time": lastexec.time,
                    "means": lastexec.mtime,
                    "var": lastexec.vtime,
                }
            )


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.pass_option_wdir
@cmd.argument("commands", nargs=-1, metavar="command args...")
def launch(commands, eid, working_dir):
    "Launch a process in background"
    pid = bmy.launch(commands, wdir=working_dir, eid=eid)

    for eid, pid in pid._d_.items():
        cmd.info("[%s]" % pid, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.pass_option_wdir
@cmd.argument("filename")
def which(filename, eid, working_dir):
    "Locate a command"
    locate = bmy.which(filename, wdir=working_dir, eid=eid)
    for eid, locate in locate._d_.items():
        if locate:
            cmd.info(locate, eid=eid)
        else:
            cmd.error("%s not found in PATH" % filename, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.option("-s", "--sig", type=types.INT, help="signal number", default=signal.SIGTERM)
@cmd.argument("pid", type=types.INT)
def kill(pid, sig, eid):
    "Kill a process on an entity"
    bmy.kill(pid, sig, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument(
    "path",
    type=types.STRING,
    required=False,
    shell_complete=completion.path_completer(only_directories=True),
)
def lsdir(path, eid):
    "List directories on the selected entity"
    dirs = bmy.lsdir(path, eid=eid)

    with cmd.EidOuput(eid=eid) as output:
        for d in dirs:
            cmd.info(d, output=output)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument(
    "path",
    type=types.STRING,
    shell_complete=completion.path_completer(only_directories=True),
)
def cd(path, eid):
    "Change the current directory on the selected entity"
    if bmy.current(eid):
        _, new = bmy.chdir(path, eid=eid)
        cmd.info(new)
    else:
        os.chdir(path)
        cmd.info(os.getcwd())


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("path", type=types.STRING, shell_complete=completion.path_completer())
def rm(path, eid):
    "Remove a file on the selected entity"
    bmy.rm(path, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="force remove directory and its contents",
)
@cmd.argument(
    "path",
    type=types.STRING,
    shell_complete=completion.path_completer(only_directories=True),
)
def rmdir(force, path, eid):
    "Remove a directory on the selected entity"
    bmy.rmdir(path, force, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument(
    "path",
    type=types.STRING,
    shell_complete=completion.path_completer(only_directories=True),
)
def mkdir(path, eid):
    "Create a directory on the selected entity"
    bmy.mkdir(path, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.option(
    "--nl",
    is_flag=True,
    help="nl: converting new line in host format, no conversion by default",
)
@cmd.option(
    "--editor",
    shell_complete=completion.local_path_executor(),
    help="editor executable path",
    default=None,
)
@cmd.argument("path", type=types.STRING, shell_complete=completion.path_completer())
def edit(path, editor, nl, eid):
    "Edit a file on the entity"
    bmy.edit(path, nl=nl, edit=editor, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("from_path", type=types.STRING, shell_complete=completion.local_path_completer())
@cmd.argument("to_path", required=False, default="", shell_complete=completion.path_completer())
def push(from_path, to_path, eid):
    "Upload file(s) to the selected entity"

    fcount = 0
    dcount = 0

    for eid, (ds, fs) in bmy.push(from_path, to_path, eid=eid)._d_.items():
        fcount = len(fs)
        dcount = len(ds)

    if dcount:
        cmd.success("%d dir%s created successfully" % (dcount, (dcount > 1) and "s" or ""))

    cmd.success("%d file%s pushed successfully" % (fcount, (fcount > 1) and "s" or ""))


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("from_path", type=types.STRING, shell_complete=completion.path_completer())
@cmd.argument(
    "to_path",
    required=False,
    default="",
    shell_complete=completion.local_path_completer(),
)
def get(from_path, to_path, eid):
    "Download file(s) from the selected entity"

    fcount = 0
    dcount = 0

    for eid, (ds, fs) in bmy.get(from_path, to_path, eid=eid)._d_.items():
        fcount = len(fs)
        dcount = len(ds)

    if dcount:
        cmd.success("%d director%s created successfully" % (dcount, (dcount > 1) and "ies" or "y"))
    cmd.success("%d file%s got successfully" % (fcount, (fcount > 1) and "s" or ""))


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("from_path", type=types.STRING, shell_complete=completion.path_completer())
@cmd.argument("to_path", type=types.STRING, shell_complete=completion.path_completer())
def cp(from_path, to_path, eid):
    "Copy file/directory on entity"
    bmy.cp(from_path, to_path, eid=eid)


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("from_path", type=types.STRING, shell_complete=completion.path_completer())
@cmd.argument("target_eid", shell_complete=completion.eids_completer)
@cmd.argument(
    "to_path",
    required=False,
    default="",
    shell_complete=completion.path_completer(use_argument_for_eid="target_eid"),
)
def transfer(from_path, to_path, eid, target_eid):
    "Download file(s) from the selected entity"

    for eid, (ds, fs) in bmy.transfer(target_eid, from_path, to_path, eid=eid)._d_.items():
        fcount = len(fs)
        dcount = len(ds)

    if dcount:
        cmd.success("%d director%s transferred successfully" % (dcount, (dcount > 1) and "ies" or "y"))
    cmd.success("%d file%s transferred successfully" % (fcount, (fcount > 1) and "s" or ""))


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("paths", nargs=-1, type=types.STRING, required=True)
def joinpath(paths, eid):
    "Join one or more path warehouse"
    cmd.info(bmy.joinpath(*paths, eid=eid))


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("path", type=types.STRING, required=True)
def realpath(path, eid):
    "Return the canonical path of the specified filename"
    cmd.info(bmy.realpath(path, eid=eid))


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("path", type=types.STRING, required=True)
def basename(path, eid):
    "Return the base name part of a path"
    cmd.info(bmy.basename(path, eid=eid))


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("path", type=types.STRING, required=True)
def dirname(path, eid):
    "Return the directory name of the specified path"
    cmd.info(bmy.dirname(path, eid=eid))


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("path", type=types.STRING, required=True, shell_complete=completion.path_completer())
def read(path, eid):
    "Display file content to console output"

    async def read():
        data = bmy.read(path, eid=eid)

        if not bmy.isgroup(data):
            data = {bmy.current(eid=eid): data}
        else:
            data = data._d_

        binding = prompt_toolkit.key_binding.KeyBindings()

        @binding.add("c-d")
        @binding.add("q")
        def _(event):
            "Quit."
            event.app.exit()

        search_field = prompt_toolkit.widgets.SearchToolbar(text_if_not_searching=[("class:not-searching", "Press '/' to start searching")])

        for _eid, data in data.items():

            def get_status_bar_text():
                return [("class:status", _eid + " - " + path)]

            text_area = prompt_toolkit.widgets.TextArea(
                text=data,
                read_only=True,
                scrollbar=True,
                line_numbers=False,
                search_field=search_field,
            )

            root_container = prompt_toolkit.layout.containers.HSplit(
                [
                    prompt_toolkit.layout.containers.Window(content=prompt_toolkit.layout.controls.FormattedTextControl(get_status_bar_text)),
                    text_area,
                    search_field,
                ]
            )

            application = prompt_toolkit.application.application.Application(
                layout=prompt_toolkit.layout.layout.Layout(root_container),
                key_bindings=binding,
                enable_page_navigation_bindings=True,
            )

            await application.run_async()

    return read


@myrrhc_cmds.command()
@cmd.pass_option_eid
@cmd.argument("path", type=types.STRING, required=True, shell_complete=completion.path_completer())
@cmd.argument("text", type=types.STRING, required=False, default=None)
def write(path, text, eid):
    "Write a text to a file"
    if text is not None:
        bmy.write(path, data=text.encode(), eid=eid)
        return

    async def write():
        text = await prompt_toolkit.shortcuts.PromptSession(
            "== %s ==\n" % path,
            multiline=True,
            bottom_toolbar="<esc> <enter> to quit",
            editing_mode=prompt_toolkit.enums.EditingMode.VI,
            enable_open_in_editor=True,
        ).prompt_async()
        cmd.info("== eof ==")
        bmy.write(path, data=text.encode(), eid=eid)

    return write



import abc
import typing

from ._ientity import (
    ISystem,
    ICoreShellService,
    ICoreFileSystemService,
    ICoreStreamService,
)
from ...provider import Stat, Protocol

__all__ = ["IRuntimeObject", "ITask", "IMyrrhOs", "IMHandle"]
__all__ += [
    "IStream",
    "IInStream",
    "IOutStream",
    "IInOutStream",
    "IFileInStream",
    "IFileOutStream",
    "IFileInOutStream",
    "Stat",
]
__all__ += [
    "IProcess",
]
__all__ += [
    "IRuntimeTaskManager",
]


class IMHandle(abc.ABC):
    @property
    @abc.abstractmethod
    def val(self) -> int:
        ...

    @property
    @abc.abstractmethod
    def closed(self) -> bool:
        ...

    @abc.abstractmethod
    def close(self) -> None:
        ...

    @abc.abstractmethod
    def detach(self) -> int:
        ...


class IFuture(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def wait(self, timeout: float) -> None:
        ...

    @classmethod
    @abc.abstractmethod
    def result(self, timeout: float) -> None:
        ...

    @classmethod
    @abc.abstractmethod
    def notify(self, msg: typing.Any) -> None:
        ...


class IRuntimeObject(abc.ABC):
    __m_ref_count__ = 0

    @property
    @abc.abstractmethod
    def ehandle(self) -> int:
        ...

    @property
    @abc.abstractmethod
    def eref(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def epath(self) -> bytes:
        ...

    @property
    @abc.abstractmethod
    def ename(self) -> bytes:
        ...

    @property
    @abc.abstractmethod
    def closed(self) -> bool:
        ...

    @abc.abstractmethod
    def close(self):
        ...

    @classmethod
    def __subclasshook__(cls, C):
        if issubclass(C, IMHandle):
            return issubclass(C.obj.__class__, cls)

        return NotImplemented


class ITask(abc.ABC):
    @abc.abstractmethod
    def task(self, future: IFuture) -> bool:
        ...

    @abc.abstractmethod
    def terminated(self) -> int | None:
        ...


class IRuntimeTaskManager(abc.ABC):
    @abc.abstractmethod
    def submit(self) -> bool:
        ...


class IMyrrhOs(ISystem, abc.ABC):
    @property
    @abc.abstractmethod
    def curdirb(self) -> bytes:
        ...

    @property
    @abc.abstractmethod
    def pardirb(self) -> bytes:
        ...

    @property
    @abc.abstractmethod
    def extsepb(self) -> bytes:
        ...

    @property
    @abc.abstractmethod
    def sepb(self) -> bytes:
        ...

    @property
    @abc.abstractmethod
    def pathsepb(self) -> bytes:
        ...

    @property
    @abc.abstractmethod
    def altsepb(self) -> bytes:
        ...

    @property
    @abc.abstractmethod
    def linesepb(self) -> bytes:
        ...

    @property
    @abc.abstractmethod
    def devnullb(self) -> bytes:
        ...

    @property
    @abc.abstractmethod
    def defpathb(self) -> bytes:
        ...

    @property
    @abc.abstractmethod
    def modules(self) -> dict[str, typing.Any]:
        ...

    @property
    @abc.abstractmethod
    def impls(self) -> dict[str, typing.Any]:
        ...

    @abc.abstractmethod
    def getbinb(self) -> dict[bytes, bytes]:
        ...

    @abc.abstractmethod
    def cwdb(self) -> bytes:
        ...

    @abc.abstractmethod
    def tmpdirb(self) -> bytes:
        ...

    @abc.abstractmethod
    def localcode(self) -> str:
        ...

    @abc.abstractmethod
    def getdefaultlocale(self) -> tuple[str]:
        ...

    @abc.abstractmethod
    def getpath(self, path: str | None = None) -> str:
        ...

    @abc.abstractmethod
    def getpathb(self, path: bytes | None = None) -> bytes:
        ...

    @abc.abstractmethod
    def envb(self) -> dict[bytes, bytes]:
        ...

    @abc.abstractmethod
    def getenvb(self, env: dict[bytes, bytes] | None = None) -> dict[bytes, bytes]:
        ...

    @abc.abstractmethod
    def getenv(self) -> dict[str, str]:
        ...

    @abc.abstractmethod
    def setenvb(self, env) -> None:
        ...

    @abc.abstractmethod
    def rdenvb(self) -> list[bytes]:
        ...

    @abc.abstractmethod
    def environkeyformat(self, key: bytes) -> bytes:
        ...

    @abc.abstractmethod
    def isabs(self, path: bytes) -> bool:
        ...

    @abc.abstractmethod
    def normpath(self, pathh: bytes) -> bytes:
        ...

    @abc.abstractmethod
    def joinpath(self, *args: list[bytes]) -> bytes:
        ...

    @abc.abstractmethod
    def basename(self, path: bytes) -> bytes:
        ...

    @abc.abstractmethod
    def dirname(self, path: bytes) -> bytes:
        ...

    @abc.abstractmethod
    def syspathb(self, path: bytes) -> bytes:
        ...

    @abc.abstractmethod
    def getshellscriptb(self, script: bytes) -> bytes:
        ...

    @abc.abstractmethod
    def getdefaultshellb(self, args: tuple[bytes, ...] = tuple()) -> bytes:
        ...

    @abc.abstractmethod
    def formatshellargs(self, args: list[bytes], *, defaultargs: list[bytes] | None = None) -> bytes:
        ...

    @abc.abstractmethod
    def sh_escape_bytes(self, string: bytes) -> bytes:
        ...

    @abc.abstractmethod
    def shellb(self) -> bytes:
        ...

    @abc.abstractmethod
    def shellargsb(self) -> list[bytes]:
        ...

    @abc.abstractmethod
    def cmd(self, cmdline: bytes, **kwargs) -> tuple[str, str, int]:
        ...

    @abc.abstractmethod
    def cmdb(self, cmdline: bytes, **kwargs) -> tuple[bytes, bytes, int]:
        ...

    @abc.abstractmethod
    def fsencoding(self) -> str:
        ...

    @abc.abstractmethod
    def fsencodeerrors(self) -> str:
        ...

    @abc.abstractmethod
    def defaultencoding(self) -> str:
        ...

    @abc.abstractmethod
    def shdecode(self, b: str | bytes, errors="surrogateescape") -> str:
        ...

    @abc.abstractmethod
    def shencode(self, s: str | bytes, errors="surrogateescape") -> bytes:
        ...

    @abc.abstractmethod
    def fsdecode(self, val: str | bytes) -> str:
        ...

    @abc.abstractmethod
    def fsencode(self, val: str | bytes) -> bytes:
        ...

    @abc.abstractmethod
    def fdcast(self, val: typing.Any, *, encoding: str | None = None) -> typing.Any:
        ...

    @abc.abstractmethod
    def fscast(self, val: typing.Any, *, encoding: str | None = None) -> typing.Any:
        ...

    @abc.abstractmethod
    def f(self, path: str | bytes | int, *, dir_fd: None | int = None) -> bytes:
        ...

    @abc.abstractmethod
    def p(self, path: str | bytes, *, dir_fd: None | int = None) -> bytes:
        ...

    @abc.abstractmethod
    def Stream(self, protocol: str | Protocol | None = None) -> ICoreStreamService:
        ...

    @abc.abstractmethod
    def Fs(self, protocol: str | Protocol | None = None) -> ICoreFileSystemService:
        ...

    @abc.abstractmethod
    def Shell(self, protocol: str | Protocol | None = None) -> ICoreShellService:
        ...


class IStream(IRuntimeObject):
    @property
    @abc.abstractmethod
    def eot(self) -> bool:
        ...

    @abc.abstractmethod
    def send_eot(self, eot: bool, id: int | None) -> bool:
        ...

    @abc.abstractmethod
    def sync(self, *, extras: dict[str, typing.Any] | None = None) -> None:
        ...

    @abc.abstractmethod
    def flush(self, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
        ...

    @abc.abstractmethod
    def stat(self, *, extras: dict[str, typing.Any] | None = None) -> Stat:
        ...


class IInStream(IStream):
    @abc.abstractmethod
    def read(self, nbytes: int | None = None, *, extras: dict[str, typing.Any] | None = None) -> bytearray:
        ...


class IOutStream(IStream):
    @abc.abstractmethod
    def write(self, data: bytes, *, extras: dict[str, typing.Any] | None = None):
        ...


class IInOutStream(IInStream, IOutStream):
    ...


class IFileInStream(IInStream):
    @abc.abstractmethod
    def seek(self, nbytes: int, *, extras: dict[str, typing.Any] | None = None) -> int:
        ...


class IFileOutStream(IOutStream):
    @abc.abstractmethod
    def truncate(self, length: int, *, extras: dict[str, typing.Any] | None = None) -> None:
        ...

    @abc.abstractmethod
    def seek(self, nbytes: int, *, extras: dict[str, typing.Any] | None = None) -> int:
        ...


class IFileInOutStream(IFileInStream, IFileOutStream, IInOutStream):
    ...


class IProcess(ITask, IRuntimeObject):
    @property
    @abc.abstractmethod
    def pid(self) -> int:
        ...

    @property
    @abc.abstractmethod
    def exit_status(self, *, extras: dict[str, typing.Any] | None = None) -> int | None:
        ...

    @abc.abstractmethod
    def terminate(self, *, extras: dict[str, typing.Any] | None = None) -> None:
        ...

    @abc.abstractmethod
    def stat(self, *, extras=None) -> Stat:
        ...



import importlib
import typing
import pydantic
import functools

from myrrh.core.services import load_ext_group

from .item import BaseItem, NoneItemType, GenericItem
from ._system import System
from ._id import Id
from ._credentials import Credentials
from ._host import Host
from ._supply import Supply, Settings


class ProviderSettings(Settings):
    model_config = pydantic.ConfigDict(
        extra="allow",
    )
    name: str

    @pydantic.field_validator("name")
    @classmethod
    def name_validate(cls, name, info):
        if name in ItemRegistry.provider_settings:
            raise ValueError(f'"{name}" is not a valid model')

        return name


class ItemRegistry:
    __single__: typing.Type["ItemRegistry"] | None = None

    warehouse_predefined_items = {
        System._type_(): System,
        Id._type_(): Id,
        Credentials._type_(): Credentials,
        Host._type_(): Host,
        NoneItemType._type_(): NoneItemType,
    }

    __warehouse_items__: dict[str, typing.Type[BaseItem]] = {}

    __warehouse_items__.update(warehouse_predefined_items)

    provider_settings: dict[str, typing.Any] = dict()

    generic = GenericItem

    def __new__(cls):
        if cls.__single__ is None:
            cls.__single__ = super().__new__(cls)
        return cls.__single__

    def __getattr__(self, name) -> typing.Type[BaseItem]:
        return self.get(name)

    def get(self, name) -> typing.Type[BaseItem]:
        try:
            return self.__warehouse_items__[name]
        except KeyError:
            pass

        raise AttributeError(f'{name} is not a valid item name, available items: {", ".join(self.items)}')

    @property
    def items(self):
        return self.__warehouse_items__

    @functools.cached_property
    def ProviderModelT(self):
        if len(self.provider_settings) > 1:
            return typing.Annotated[typing.Union[*self.provider_settings.values()], pydantic.Field(discriminator="name")] | ProviderSettings  # type: ignore

        if len(self.provider_settings):
            return [p for p in self.provider_settings.values()][0] | ProviderSettings

        return ProviderSettings

    @functools.cached_property
    def WarehouseItemT(self):
        return typing.Annotated[typing.Union[*self.warehouse_predefined_items.values()], pydantic.Field(discriminator="type_")] | GenericItem  # type: ignore

    @property
    def FactorySupply(self) -> typing.Type[Supply]:
        return Supply[self.ProviderModelT]  # type: ignore[name-defined]

    def warehouse_model_validate(self, *a, **kwa):
        cls = pydantic.TypeAdapter(self.WarehouseItemT).validate_python(*a, **kwa)
        return cls

    def provider_model_validate(self, *a, **kwa):
        cls = pydantic.TypeAdapter(self.ProviderModelT).validate_python(*a, **kwa)
        return cls

    def register_warehouse(self, item_cls):
        self.warehouse_predefined_items[item_cls._type_()] = item_cls
        self.__warehouse_items__[item_cls._type_()] = item_cls
        try:
            del self.WarehouseItemT
        except AttributeError:
            pass

    def register_provider_model(self, provider_model: ProviderSettings):
        for value in typing.get_args(provider_model.model_fields["name"].annotation):
            self.provider_settings[value] = provider_model
        try:
            del self.ProviderModelT
        except AttributeError:
            pass


def register_warehouse(module_name: str, item: str):
    module = importlib.import_module(module_name)
    item_cls = getattr(module, item)
    return ItemRegistry().register_warehouse(item_cls)


def register_provider_model(module_name: str, model: str):
    module = importlib.import_module(module_name)
    model_cls = getattr(module, model)
    return ItemRegistry().register_provider_model(model_cls)


load_ext_group("myrrh.warehouse.registry")
